##
# Copyright 2020-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing LLVM, implemented as an easyblock

@author: Dmitri Gribenko (National Technical University of Ukraine "KPI")
@author: Ward Poelmans (Ghent University)
@author: Alan O'Cais (Juelich Supercomputing Centre)
@author: Maxime Boissonneault (Digital Research Alliance of Canada, Universite Laval)
@author: Simon Branford (University of Birmingham)
@author: Kenneth Hoste (Ghent University)
@author: Davide Grassano (CECAM HQ - Lausanne)
"""
import contextlib
import glob
import os
import re
import stat

from easybuild.framework.easyconfig import CUSTOM
from easybuild.toolchains.compiler.clang import Clang
from easybuild.tools import LooseVersion
from easybuild.tools.utilities import trace_msg
from easybuild.tools.build_log import EasyBuildError, print_msg
from easybuild.tools.config import ERROR, IGNORE, SEARCH_PATH_LIB_DIRS, build_option
from easybuild.tools.environment import setvar
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, copy_dir, adjust_permissions
from easybuild.tools.filetools import mkdir, remove_file, symlink, which, write_file, remove_dir
from easybuild.tools.modules import MODULE_LOAD_ENV_HEADERS, get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd, EasyBuildExit
from easybuild.tools.systemtools import AARCH32, AARCH64, POWER, RISCV64, X86_64, POWER_LE
from easybuild.tools.systemtools import get_cpu_architecture, get_cpu_family, get_shared_lib_ext

from easybuild.easyblocks.generic.cmakemake import CMakeMake, get_cmake_python_config_dict

BUILD_TARGET_AMDGPU = 'AMDGPU'
BUILD_TARGET_NVPTX = 'NVPTX'

LLVM_TARGETS = [
    'AArch64', BUILD_TARGET_AMDGPU, 'ARM', 'AVR', 'BPF', 'Hexagon', 'Lanai', 'LoongArch', 'Mips', 'MSP430',
    BUILD_TARGET_NVPTX, 'PowerPC', 'RISCV', 'Sparc', 'SPIRV', 'SystemZ', 'VE', 'WebAssembly', 'X86', 'XCore',
    'all'
]
LLVM_EXPERIMENTAL_TARGETS = [
    'ARC', 'CSKY', 'DirectX', 'M68k', 'Xtensa',
]
ALL_TARGETS = LLVM_TARGETS + LLVM_EXPERIMENTAL_TARGETS

DEFAULT_TARGETS_MAP = {
    AARCH32: ['ARM'],
    AARCH64: ['AArch64'],
    POWER: ['PowerPC'],
    RISCV64: ['RISCV'],
    X86_64: ['X86'],
}

remove_gcc_dependency_opts = {
    'CLANG_DEFAULT_CXX_STDLIB': 'libc++',
    'CLANG_DEFAULT_RTLIB': 'compiler-rt',
    # Moved to general_opts for ease of building with openmp offload (or other multi-stage builds)
    # 'CLANG_DEFAULT_LINKER': 'lld',
    'CLANG_DEFAULT_UNWINDLIB': 'libunwind',

    'COMPILER_RT_BUILD_GWP_ASAN': 'Off',
    'COMPILER_RT_ENABLE_INTERNAL_SYMBOLIZER': 'On',
    'COMPILER_RT_ENABLE_STATIC_UNWINDER': 'On',  # https://lists.llvm.org/pipermail/llvm-bugs/2016-July/048424.html
    'COMPILER_RT_USE_BUILTINS_LIBRARY': 'On',
    'COMPILER_RT_USE_LIBCXX': 'On',
    'COMPILER_RT_USE_LLVM_UNWINDER': 'On',

    'LIBCXX_CXX_ABI': 'libcxxabi',
    'LIBCXX_DEFAULT_ABI_LIBRARY': 'libcxxabi',
    # Needed as libatomic could not be present on the system (compilation and tests will succeed because of the
    # GCCcore builddep, but usage/sanity check will fail due to missing libatomic)
    'LIBCXX_HAS_ATOMIC_LIB': 'NO',
    'LIBCXX_HAS_GCC_S_LIB': 'Off',
    'LIBCXX_USE_COMPILER_RT': 'On',

    'LIBCXXABI_HAS_GCC_S_LIB': 'Off',
    'LIBCXXABI_USE_LLVM_UNWINDER': 'On',
    'LIBCXXABI_USE_COMPILER_RT': 'On',

    'LIBUNWIND_HAS_GCC_S_LIB': 'Off',
    'LIBUNWIND_USE_COMPILER_RT': 'On',

    # Libxml2 from system gets automatically detected and linked in bringing dependencies from stdc++, gcc_s, icuuc, etc
    # Moved to a check at the configure step. See https://github.com/easybuilders/easybuild-easyconfigs/issues/22491
    # 'LLVM_ENABLE_LIBXML2': 'Off',

    'SANITIZER_USE_STATIC_LLVM_UNWINDER': 'On',
}

disable_werror = {
    'BENCHMARK_ENABLE_WERROR': 'Off',
    'COMPILER_RT_ENABLE_WERROR': 'Off',
    'FLANG_ENABLE_WERROR': 'Off',
    'LIBC_WNO_ERROR': 'On',
    'LIBCXX_ENABLE_WERROR': 'Off',
    'LIBUNWIND_ENABLE_WERROR': 'Off',
    'LLVM_ENABLE_WERROR': 'Off',
    'OPENMP_ENABLE_WERROR': 'Off',
}

general_opts = {
    'CMAKE_VERBOSE_MAKEFILE': 'ON',
    'LLVM_INCLUDE_BENCHMARKS': 'OFF',
    'LLVM_INSTALL_UTILS': 'ON',
    # If EB is launched from a venv, avoid giving priority to the venv's python
    'Python3_FIND_VIRTUALENV': 'STANDARD',
}


@contextlib.contextmanager
def _wrap_env(path="", ld_path=""):
    """Wrap the environment with $PATH and $LD_LIBRARY_PATH."""
    orig_path = os.getenv('PATH', '')
    orig_ld_library_path = os.getenv('LD_LIBRARY_PATH', '')

    path = ':'.join(filter(None, [path, orig_path]))
    ld_path = ':'.join(filter(None, [ld_path, orig_ld_library_path]))

    setvar('PATH', path)
    setvar('LD_LIBRARY_PATH', ld_path)

    try:
        yield
    finally:
        setvar('PATH', orig_path)
        setvar('LD_LIBRARY_PATH', orig_ld_library_path)


def get_arch_prefix():
    """Return the architecture prefix"""
    arch = get_cpu_architecture()
    if arch == POWER:
        if get_cpu_family() == POWER_LE:
            return 'powerpc64le'
        else:
            return 'powerpc64'
    else:
        return arch.lower()


class EB_LLVM(CMakeMake):
    """
    Support for building and installing LLVM
    """

    minimal_conflicts = [
        'build_bolt',
        'build_clang_extras',
        'build_lld',
        'build_lldb',
        'build_openmp',
        'build_openmp_tools',
        'build_runtimes',
        'bootstrap',
        'full_llvm',
        'python_bindings',
        'usepolly',
    ]

    # Create symlink between equivalent host triples, useful so that other build processes that relies on older
    # triple names can still work when passing the old name to --target
    symlink_lst = [
        ('x86_64-unknown-linux-gnu', 'x86_64-pc-linux'),
        ('x86_64-unknown-linux-gnu', 'x86_64-pc-linux-gnu'),
    ]

    # From LLVM 19, GCC_INSTALL_PREFIX is not supported anymore to hardcode the GCC installation path into the binaries;
    # Now every compilers needs a .cfg file with the --gcc-install-dir option
    # This list tells which compilers need to have a .cfg file created
    # NOTE: flang is the expected name also for the 'flang-new' compiler
    cfg_compilers = ['clang', 'clang++', 'flang']

    @staticmethod
    def extra_options():
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'amd_gfx_list': [None, "List of AMDGPU targets to build for.", CUSTOM],
            'assertions': [False, "Enable assertions.  Helps to catch bugs in Clang.", CUSTOM],
            'bootstrap': [True, "Build LLVM-Clang using itself", CUSTOM],
            'build_bolt': [False, "Build the LLVM bolt binary optimizer", CUSTOM],
            'build_clang_extras': [False, "Build the LLVM Clang extra tools", CUSTOM],
            'build_lld': [False, "Build the LLVM lld linker", CUSTOM],
            'build_lldb': [False, "Build the LLVM lldb debugger", CUSTOM],
            'build_openmp': [True, "Build the LLVM OpenMP runtime", CUSTOM],
            'build_openmp_offload': [True, "Build the LLVM OpenMP offload runtime", CUSTOM],
            'build_openmp_tools': [True, "Build the LLVM OpenMP tools interface", CUSTOM],
            'build_runtimes': [False, "Build the LLVM runtimes (compiler-rt, libunwind, libcxx, libcxxabi)", CUSTOM],
            'build_targets': [None, "Build targets for LLVM (host architecture if None). Possible values: " +
                                    ', '.join(ALL_TARGETS), CUSTOM],
            'debug_tests': [True, "Enable verbose output for tests", CUSTOM],
            'disable_werror': [False, "Disable -Werror for all projects", CUSTOM],
            'enable_rtti': [True, "Enable RTTI", CUSTOM],
            'full_llvm': [False, "Build LLVM without any dependency", CUSTOM],
            'minimal': [False, "Build LLVM only", CUSTOM],
            'python_bindings': [False, "Install python bindings", CUSTOM],
            'skip_all_tests': [False, "Skip running of tests", CUSTOM],
            'skip_sanitizer_tests': [True, "Do not run the sanitizer tests", CUSTOM],
            'test_suite_ignore_patterns': [None, "List of test to ignore (if the string matches)", CUSTOM],
            'test_suite_max_failed': [0, "Maximum number of failing tests (does not count allowed failures)", CUSTOM],
            'test_suite_timeout_single': [None, "Timeout for each individual test in the test suite", CUSTOM],
            'test_suite_timeout_total': [None, "Timeout for total running time of the testsuite", CUSTOM],
            'use_pic': [True, "Build with Position Independent Code (PIC)", CUSTOM],
            'usepolly': [False, "Build Clang with polly", CUSTOM],
        })

        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialize LLVM-specific variables."""
        super().__init__(*args, **kwargs)

        self.llvm_src_dir = None
        self.llvm_obj_dir_stage1 = None
        self.llvm_obj_dir_stage2 = None
        self.llvm_obj_dir_stage3 = None
        self.intermediate_projects = ['llvm', 'clang']
        self.intermediate_runtimes = ['compiler-rt', 'libunwind', 'libcxx', 'libcxxabi']
        if not self.cfg['minimal']:
            self.final_projects = ['llvm', 'mlir', 'clang', 'flang']
        else:
            self.final_projects = ['llvm']
        self.final_runtimes = []
        self.gcc_prefix = None
        self.runtimes_cmake_args = {
            'CMAKE_C_COMPILER': [],
            'CMAKE_C_FLAGS': [],
            'CMAKE_CXX_COMPILER': [],
            'CMAKE_CXX_FLAGS': [],
            'CMAKE_EXE_LINKER_FLAGS': [],
            }
        self.offload_targets = ['host']
        # self._added_librt = None
        self.host_triple = None

        # Shared
        off_opts, on_opts = [], []
        self.build_shared = self.cfg.get('build_shared_libs', False)
        if self.build_shared:
            self.cfg['build_shared_libs'] = None
            on_opts.extend(['LLVM_BUILD_LLVM_DYLIB', 'LLVM_LINK_LLVM_DYLIB', 'LIBCXX_ENABLE_SHARED',
                            'LIBCXXABI_ENABLE_SHARED', 'LIBUNWIND_ENABLE_SHARED'])
        else:
            off_opts.extend(['LIBCXX_ENABLE_ABI_LINKER_SCRIPT', 'LIBCXX_ENABLE_SHARED', 'LIBCXXABI_ENABLE_SHARED',
                             'LIBUNWIND_ENABLE_SHARED', 'LLVM_BUILD_LLVM_DYLIB', 'LLVM_LINK_LLVM_DYLIB'])
            on_opts.extend(['LIBCXX_ENABLE_STATIC', 'LIBCXX_ENABLE_STATIC_ABI_LIBRARY', 'LIBCXXABI_ENABLE_STATIC',
                            'LIBUNWIND_ENABLE_STATIC'])

        # RTTI
        if self.cfg['enable_rtti']:
            on_opts.extend(['LLVM_ENABLE_RTTI', 'LLVM_REQUIRES_RTTI'])
            # Does not work yet with Flang
            # on_opts.append('LLVM_ENABLE_EH')

        if self.cfg['use_pic']:
            on_opts.append('CMAKE_POSITION_INDEPENDENT_CODE')

        for opt in on_opts:
            general_opts[opt] = 'ON'

        for opt in off_opts:
            general_opts[opt] = 'OFF'

        self.full_llvm = self.cfg['full_llvm']

        if self.cfg['minimal']:
            conflicts = [_ for _ in self.minimal_conflicts if self.cfg[_]]
            if conflicts:
                raise EasyBuildError("Minimal build conflicts with '%s'", ', '.join(conflicts))

        # Other custom options
        if self.full_llvm:
            if not self.cfg['bootstrap']:
                raise EasyBuildError("Full LLVM build requires bootstrap build")
            if not self.cfg['build_lld']:
                raise EasyBuildError("Full LLVM build requires building lld")
            if not self.cfg['build_runtimes']:
                raise EasyBuildError("Full LLVM build requires building runtimes")
            self.log.info("Building LLVM without any GCC dependency")

        if self.cfg['disable_werror']:
            general_opts.update(disable_werror)

        if self.cfg['build_runtimes']:
            self.final_runtimes += ['compiler-rt', 'libunwind', 'libcxx', 'libcxxabi']

        if self.cfg['build_openmp']:
            self.final_projects.append('openmp')

        if self.cfg['build_openmp_offload']:
            if not self.cfg['build_openmp']:
                raise EasyBuildError("Building OpenMP offload requires building OpenMP runtime")
            # LLVM 19 added a new runtime target for explicit offloading
            # https://discourse.llvm.org/t/llvm-19-1-0-no-library-libomptarget-nvptx-sm-80-bc-found/81343
            if LooseVersion(self.version) >= LooseVersion('19'):
                self.log.debug("Explicitly enabling OpenMP offloading for LLVM >= 19")
                self.final_runtimes.append('offload')
            else:
                self.log.warning("OpenMP offloading is included with the OpenMP runtime for LLVM < 19")

        if self.cfg['build_openmp_tools']:
            if not self.cfg['build_openmp']:
                raise EasyBuildError("Building OpenMP tools requires building OpenMP runtime")

        if self.cfg['usepolly']:
            self.final_projects.append('polly')

        if self.cfg['build_clang_extras']:
            self.final_projects.append('clang-tools-extra')

        if self.cfg['build_lld']:
            self.intermediate_projects.append('lld')
            self.final_projects.append('lld')
            # This should be the default to make offload multi-stage compilations easier
            general_opts['CLANG_DEFAULT_LINKER'] = 'lld'
            general_opts['FLANG_DEFAULT_LINKER'] = 'lld'

        if self.cfg['build_lldb']:
            self.final_projects.append('lldb')
            if self.full_llvm:
                remove_gcc_dependency_opts['LLDB_ENABLE_LIBXML2'] = 'Off'
                remove_gcc_dependency_opts['LLDB_ENABLE_LZMA'] = 'Off'
                remove_gcc_dependency_opts['LLDB_ENABLE_PYTHON'] = 'Off'

        if self.cfg['build_bolt']:
            self.final_projects.append('bolt')

        # Fix for https://github.com/easybuilders/easybuild-easyblocks/issues/3689
        if LooseVersion(self.version) < LooseVersion('16'):
            general_opts['LLVM_INCLUDE_GO_TESTS'] = 'OFF'

        # Sysroot
        sysroot = build_option('sysroot')
        if sysroot:
            general_opts['DEFAULT_SYSROOT'] = sysroot
            general_opts['CMAKE_SYSROOT'] = sysroot

        # list of CUDA compute capabilities to use can be specifed in two ways (where (2) overrules (1)):
        # (1) in the easyconfig file, via the custom cuda_compute_capabilities;
        # (2) in the EasyBuild configuration, via --cuda-compute-capabilities configuration option;
        cuda_cc_list = build_option('cuda_compute_capabilities') or self.cfg['cuda_compute_capabilities'] or []
        cuda_toolchain = hasattr(self.toolchain, 'COMPILER_CUDA_FAMILY')
        amd_gfx_list = self.cfg['amd_gfx_list'] or []

        # List of (lower-case) dependencies
        self.deps = [dep['name'].lower() for dep in self.cfg.dependencies()]

        # Build targets
        build_targets = self.cfg['build_targets'] or []
        if not build_targets:
            self.log.debug("No build targets specified, using default detection")
            arch = get_cpu_architecture()
            if arch not in DEFAULT_TARGETS_MAP:
                raise EasyBuildError("No default build targets defined for CPU architecture %s.", arch)
            build_targets += DEFAULT_TARGETS_MAP[arch]

            # If CUDA is included as a dep, add NVPTX as a target
            # There are (old) toolchains with CUDA as part of the toolchain
            if 'cuda' in self.deps or cuda_toolchain:
                self.log.info("CUDA dependency detected, adding NVPTX as a target")
                build_targets.append(BUILD_TARGET_NVPTX)
            elif cuda_cc_list:
                self.log.info("CUDA compute capabilities specified, adding NVPTX as a target")
                build_targets.append(BUILD_TARGET_NVPTX)

            # For AMDGPU support during runtime we need ROCR-Runtime and ROCT-Thunk-Interface. While split into
            # separate packages pre ROCm 6.2, it is now combined into ROCR-Runtime. As ROCR-Thunk-Interface was a
            # dependency for ROCR-Runtime before, checking for ROCR-Runtime as a dependency is sufficient.
            # Generally, ROCR-Runtime is not a hard dependency for LLVM. If not found, LLVM can still build
            # an offload-capable compiler runtime, and will try to dlopen the required libraries at runtime.
            # Therefore, also allow the build without ROCR-Runtime, with only the desired architecture list being set.
            # https://openmp.llvm.org/SupportAndFAQ.html#q-how-to-build-an-openmp-amdgpu-offload-capable-compiler
            if 'rocr-runtime' in self.deps:
                self.log.info("ROCR-Runtime dependency detected, adding AMDGPU as a target")
                build_targets.append(BUILD_TARGET_AMDGPU)
            elif amd_gfx_list:
                self.log.info("AMD GPU list specified, adding AMDGPU as a target")
                build_targets.append(BUILD_TARGET_AMDGPU)

            self.cfg['build_targets'] = build_targets
            self.log.debug("Using %s as default build targets for CPU architecture %s.", build_targets, arch)

        unknown_targets = set(build_targets) - set(ALL_TARGETS)

        if unknown_targets:
            raise EasyBuildError("Some of the chosen build targets (%s) are not in %s.",
                                 ', '.join(unknown_targets), ', '.join(ALL_TARGETS))
        exp_targets = set(build_targets) & set(LLVM_EXPERIMENTAL_TARGETS)
        if exp_targets:
            self.log.warning("Experimental targets %s are being used.", ', '.join(exp_targets))

        all_target_cond = 'all' in build_targets
        self.nvptx_target_cond = (BUILD_TARGET_NVPTX in build_targets) or all_target_cond
        self.amdgpu_target_cond = (BUILD_TARGET_AMDGPU in build_targets) or all_target_cond

        if ('cuda' in self.deps or cuda_toolchain) and not self.nvptx_target_cond:
            raise EasyBuildError("CUDA dependency detected, but NVPTX not in manually specified build targets")
        if cuda_cc_list and not self.nvptx_target_cond:
            raise EasyBuildError(
                "CUDA compute capabilities specified, but NVPTX not in manually specified build targets"
            )
        if 'rocr-runtime' in self.deps and not self.amdgpu_target_cond:
            raise EasyBuildError(
                "ROCR-Runtime dependency detected, but AMDGPU not in manually specified build targets"
            )
        if amd_gfx_list and not self.amdgpu_target_cond:
            raise EasyBuildError("AMD GPU list specified, but AMDGPU not in manually specified build targets")

        self.build_targets = build_targets or []

        # Enable offload targets for LLVM >= 18
        self.cuda_cc = []
        self.amd_gfx = []
        if self.cfg['build_openmp_offload'] and LooseVersion(self.version) >= LooseVersion('18'):
            if self.nvptx_target_cond:
                if LooseVersion(self.version) < LooseVersion('20') and not cuda_cc_list:
                    raise EasyBuildError(
                        f"LLVM < 20 requires 'cuda_compute_capabilities' to build with {BUILD_TARGET_NVPTX}"
                    )
                self.cuda_cc = [cc.replace('.', '') for cc in cuda_cc_list]
                self.offload_targets += ['cuda']
                self.log.debug("Enabling `cuda` offload target")
            if self.amdgpu_target_cond:
                if LooseVersion(self.version) < LooseVersion('20') and not amd_gfx_list:
                    raise EasyBuildError(f"LLVM < 20 requires 'amd_gfx_list' to build with {BUILD_TARGET_AMDGPU}")
                self.amd_gfx = amd_gfx_list
                self.offload_targets += ['amdgpu']  # Used for LLVM >= 19
                self.log.debug("Enabling `amdgpu` offload target")

        general_opts['CMAKE_BUILD_TYPE'] = self.build_type
        general_opts['LLVM_TARGETS_TO_BUILD'] = '"%s"' % ';'.join(build_targets)

        self._cmakeopts = {}
        self._cfgopts = list(filter(None, self.cfg.get('configopts', '').split()))

    def prepare_step(self, *args, **kwargs):
        """Prepare step, modified to ensure install dir is deleted before building"""
        super(EB_LLVM, self).prepare_step(*args, **kwargs)
        # re-create installation dir (deletes old installation),
        # Needed to ensure hardcoded rpath do not point to old installation during runtime builds and testing
        self.make_installdir()

    def _add_cmake_runtime_args(self):
        """Generate the value for 'RUNTIMES_CMAKE_ARGS' and add it to the cmake options."""
        if self.runtimes_cmake_args:
            args = []
            for key, val in self.runtimes_cmake_args.items():
                if isinstance(val, list):
                    val = ' '.join(val)
                if val:
                    args.append('-D%s=%s' % (key, val))
            self._cmakeopts['RUNTIMES_CMAKE_ARGS'] = '"%s"' % ';'.join(args)

    def _configure_general_build(self):
        """General configuration step for LLVM."""
        self._cmakeopts.update(general_opts)
        self._add_cmake_runtime_args()

    def _configure_intermediate_build(self):
        """Configure the intermediate stages of the build."""
        self._cmakeopts['LLVM_ENABLE_PROJECTS'] = '"%s"' % ';'.join(self.intermediate_projects)
        self._cmakeopts['LLVM_ENABLE_RUNTIMES'] = '"%s"' % ';'.join(self.intermediate_runtimes)

    def _configure_final_build(self):
        """Configure the final stage of the build."""
        self._cmakeopts['LLVM_ENABLE_PROJECTS'] = '"%s"' % ';'.join(self.final_projects)
        self._cmakeopts['LLVM_ENABLE_RUNTIMES'] = '"%s"' % ';'.join(self.final_runtimes)

        hwloc_root = get_software_root('hwloc')
        if hwloc_root:
            self.log.info("Using %s as hwloc root", hwloc_root)
            self._cmakeopts['LIBOMP_USE_HWLOC'] = 'ON'
            self._cmakeopts['LIBOMP_HWLOC_INSTALL_DIR'] = hwloc_root

        if 'openmp' in self.final_projects:
            if LooseVersion(self.version) >= LooseVersion('19') and self.cfg['build_openmp_offload']:
                self.runtimes_cmake_args['LIBOMPTARGET_PLUGINS_TO_BUILD'] = '%s' % '|'.join(self.offload_targets)
            self._cmakeopts['OPENMP_ENABLE_LIBOMPTARGET'] = 'ON'
            self._cmakeopts['LIBOMP_INSTALL_ALIASES'] = 'OFF'
            if not self.cfg['build_openmp_tools']:
                self._cmakeopts['OPENMP_ENABLE_OMPT_TOOLS'] = 'OFF'

        # Make sure tests are not running with more than 'parallel' tasks
        parallel = self.cfg.parallel
        if not build_option('mpi_tests'):
            parallel = 1
        lit_args = [f'-j {parallel}']
        if self.cfg['debug_tests']:
            lit_args += ['-v']
        timeout_single = self.cfg['test_suite_timeout_single']
        if timeout_single:
            lit_args += ['--timeout', str(timeout_single)]
        timeout_total = self.cfg['test_suite_timeout_total']
        if timeout_total:
            lit_args += ['--max-time', str(timeout_total)]
        self._cmakeopts['LLVM_LIT_ARGS'] = '"%s"' % ' '.join(lit_args)

        if self.cfg['usepolly']:
            self._cmakeopts['LLVM_POLLY_LINK_INTO_TOOLS'] = 'ON'
        if not self.cfg['skip_all_tests']:
            self._cmakeopts['LLVM_INCLUDE_TESTS'] = 'ON'
            self._cmakeopts['LLVM_BUILD_TESTS'] = 'ON'

    @staticmethod
    def _get_gcc_prefix():
        """Get the GCC prefix for the build."""
        arch = get_arch_prefix()
        gcc_root = get_software_root('GCCcore')
        gcc_version = get_software_version('GCCcore')
        # If that doesn't work, try with GCC
        if gcc_root is None:
            gcc_root = get_software_root('GCC')
            gcc_version = get_software_version('GCC')
        # If that doesn't work either, print error and exit
        if gcc_root is None:
            raise EasyBuildError("Can't find GCC or GCCcore to use")

        pattern = os.path.join(gcc_root, 'lib', 'gcc', f'{arch}-*', gcc_version)
        matches = glob.glob(pattern)
        if not matches:
            raise EasyBuildError("Can't find GCC version %s for architecture %s in %s", gcc_version, arch, pattern)
        gcc_prefix = os.path.abspath(matches[0])

        return gcc_root, gcc_prefix

    def _set_gcc_prefix(self):
        """Set the GCC prefix for the build."""
        if self.gcc_prefix is None:
            gcc_root, gcc_prefix = self._get_gcc_prefix()

            # --gcc-toolchain and --gcc-install-dir for flang are not supported before LLVM 19
            # https://github.com/llvm/llvm-project/pull/87360
            if LooseVersion(self.version) < LooseVersion('19'):
                self.log.debug("Using GCC_INSTALL_PREFIX")
                general_opts['GCC_INSTALL_PREFIX'] = gcc_root
            else:
                # See https://github.com/llvm/llvm-project/pull/85891#issuecomment-2021370667
                self.log.debug("Using '--gcc-install-dir' in CMAKE_C_FLAGS and CMAKE_CXX_FLAGS")
                self.runtimes_cmake_args['CMAKE_C_FLAGS'] += ['--gcc-install-dir=%s' % gcc_prefix]
                self.runtimes_cmake_args['CMAKE_CXX_FLAGS'] += ['--gcc-install-dir=%s' % gcc_prefix]

            self.gcc_prefix = gcc_prefix
        self.log.debug("Using %s as the gcc install location", self.gcc_prefix)

    def configure_step(self):
        """
        Install extra tools in bin/; enable zlib if it is a dep; optionally enable rtti; and set the build target
        """
        # Allow running with older versions of LLVM for minimal builds in order to replace EB_LLVM easyblock
        if not self.cfg['minimal'] and LooseVersion(self.version) < LooseVersion('18.1.6'):
            raise EasyBuildError("LLVM version %s is not supported, please use version 18.1.6 or newer", self.version)

        # Allow running with older versions of LLVM for minimal builds in order to replace EB_LLVM easyblock
        gcc_version = get_software_version('GCCcore')
        if not self.cfg['minimal'] and LooseVersion(gcc_version) < LooseVersion('13'):
            raise EasyBuildError("LLVM %s requires GCC 13 or newer, found %s", self.version, gcc_version)

        # Lit is needed for running tests-suite
        lit_root = get_software_root('lit')
        if not lit_root:
            if not self.cfg['skip_all_tests']:
                raise EasyBuildError("Can't find 'lit', needed for running tests-suite")

        timeouts = self.cfg['test_suite_timeout_single'] or self.cfg['test_suite_timeout_total']
        if not self.cfg['skip_all_tests'] and timeouts:
            psutil_root = get_software_root('psutil')
            if not psutil_root:
                raise EasyBuildError("Can't find 'psutil', needed for running tests-suite with timeout")

        # Parallel build
        self.make_parallel_opts = ""
        if self.cfg.parallel:
            self.make_parallel_opts = f"-j {self.cfg.parallel}"

        # CMAKE_INSTALL_PREFIX and LLVM start directory are set here instead of in __init__ to
        # ensure this easyblock can be used as a Bundle component, see
        # https://github.com/easybuilders/easybuild-easyblocks/issues/3680
        general_opts['CMAKE_INSTALL_PREFIX'] = self.installdir
        start_dir = self.cfg['start_dir']
        if start_dir:
            self.llvm_src_dir = os.path.join(self.builddir, start_dir)
            self.log.debug("Using `%s` as the source dir from start_dir", self.llvm_src_dir)
        else:
            self.llvm_src_dir = os.path.join(self.builddir, 'llvm-project-%s.src' % self.version)
            self.log.debug("Using `%s` as the source dir from easyblock", self.llvm_src_dir)

        # Bootstrap
        self.llvm_obj_dir_stage1 = os.path.join(self.builddir, 'llvm.obj.1')
        if self.cfg['bootstrap']:
            self.log.info("Initialising for bootstrap build.")
            self.llvm_obj_dir_stage2 = os.path.join(self.builddir, 'llvm.obj.2')
            self.llvm_obj_dir_stage3 = os.path.join(self.builddir, 'llvm.obj.3')
            self.final_dir = self.llvm_obj_dir_stage3
            mkdir(self.llvm_obj_dir_stage2)
            mkdir(self.llvm_obj_dir_stage3)
        else:
            self.log.info("Initialising for single stage build.")
            self.final_dir = self.llvm_obj_dir_stage1

        general_opts['LLVM_ENABLE_ASSERTIONS'] = 'ON' if self.cfg['assertions'] else 'OFF'

        # Dependencies based persistent options (should be reused across stages)
        # Libxml2
        xml2_root = get_software_root('libxml2')
        # Explicitly disable libxml2 if not found to avoid linking against system libxml2
        if xml2_root:
            if self.full_llvm:
                self.log.warning("LLVM is being built in 'full_llvm' mode, libxml2 will not be used")
                general_opts['LLVM_ENABLE_LIBXML2'] = 'OFF'
            else:
                general_opts['LLVM_ENABLE_LIBXML2'] = 'ON'
        else:
            general_opts['LLVM_ENABLE_LIBXML2'] = 'OFF'

        # If 'ON', risk finding a system zlib or zstd leading to including /usr/include as -isystem that can lead
        # to errors during compilation of 'offload.tools.kernelreplay' due to the inclusion of LLVMSupport (19.x)
        general_opts['LLVM_ENABLE_ZLIB'] = 'ON' if get_software_root('zlib') else 'OFF'
        general_opts['LLVM_ENABLE_ZSTD'] = 'ON' if get_software_root('zstd') else 'OFF'
        # Should not use system SWIG if present
        general_opts['LLDB_ENABLE_SWIG'] = 'ON' if get_software_root('SWIG') else 'OFF'

        z3_root = get_software_root("Z3")
        if z3_root:
            self.log.info("Using %s as Z3 root", z3_root)
            general_opts['LLVM_ENABLE_Z3_SOLVER'] = 'ON'
            general_opts['LLVM_Z3_INSTALL_DIR'] = z3_root
        else:
            general_opts['LLVM_ENABLE_Z3_SOLVER'] = 'OFF'

        python_opts = get_cmake_python_config_dict()
        general_opts.update(python_opts)
        self.runtimes_cmake_args.update(python_opts)

        if self.cfg['bootstrap']:
            self._configure_intermediate_build()
        else:
            self._configure_final_build()

        if self.cfg['skip_sanitizer_tests'] and build_option('strict') != ERROR:
            self.log.info("Disabling the sanitizer tests")
            self.disable_sanitizer_tests()

        # Remove python bindings tests causing uncaught exception in the build
        cmakelists_tests = os.path.join(self.llvm_src_dir, 'clang', 'CMakeLists.txt')
        regex_subs = []
        regex_subs.append((r'add_subdirectory\(bindings/python/tests\)', ''))
        apply_regex_substitutions(cmakelists_tests, regex_subs)

        self._set_gcc_prefix()

        # If we don't want to build with CUDA (not in dependencies) trick CMakes FindCUDA module into not finding it by
        # using the environment variable which is used as-is and later checked for a falsy value when determining
        # whether CUDA was found
        if not get_software_root('CUDA'):
            setvar('CUDA_NVCC_EXECUTABLE', 'IGNORE')

        if self.cfg['build_openmp_offload'] and LooseVersion('19') <= LooseVersion(self.version) < LooseVersion('20'):
            gpu_archs = []
            gpu_archs += ['sm_%s' % cc for cc in self.cuda_cc]
            gpu_archs += self.amd_gfx
            if gpu_archs:
                self.runtimes_cmake_args['LIBOMPTARGET_DEVICE_ARCHITECTURES'] = '%s' % '|'.join(gpu_archs)

        self._configure_general_build()
        self.add_cmake_opts()

        src_dir = os.path.join(self.llvm_src_dir, 'llvm')
        output = super().configure_step(builddir=self.llvm_obj_dir_stage1, srcdir=src_dir)

        # Get LLVM_HOST_TRIPLE (e.g. x86_64-unknown-linux-gnu) from the output
        for line in output.splitlines():
            if 'llvm host triple' in line.lower():
                self.host_triple = line.split(':')[1].strip()
                break
        else:
            # LLVM_HOST_TRIPLE needs to be set when building runtimes or bootstrapping.
            if self.cfg['build_runtimes'] or self.cfg['bootstrap']:
                raise EasyBuildError("`LLVM_HOST_TRIPLE` not found in the output of the configure step")
            # Otherwise it can be inferred a posteriori from the install directory
            else:
                self.log.warning("`LLVM_HOST_TRIPLE` not found in the output of the configure step")

        if not self.cfg['bootstrap'] and build_option('rpath') and self._cmakeopts['LLVM_ENABLE_RUNTIMES'] != '""':
            # Ensure RPATH wrappers are used for the runtimes also at the first stage
            # Call configure again now that the host triple is known from the previous configure call
            remove_dir(self.llvm_obj_dir_stage1)
            self._prepare_runtimes_rpath_wrappers(self.llvm_obj_dir_stage1)
            self.add_cmake_opts()
            trace_msg("Reconfiguring LLVM to use the RPATH wrappers for the runtimes")
            super(EB_LLVM, self).configure_step(builddir=self.llvm_obj_dir_stage1, srcdir=src_dir)

    def disable_sanitizer_tests(self):
        """Disable the tests of all the sanitizers by removing the test directories from the build system"""
        cmakelists_tests = os.path.join(self.llvm_src_dir, 'compiler-rt', 'test', 'CMakeLists.txt')
        regex_subs = [(r'compiler_rt_test_runtime.*san.*', '')]
        apply_regex_substitutions(cmakelists_tests, regex_subs)

    def add_cmake_opts(self):
        """Add LLVM-specific CMake options."""
        base_opts = self._cfgopts.copy()
        for k, v in self._cmakeopts.items():
            base_opts.append('-D%s=%s' % (k, v))
        self.cfg['configopts'] = ' '.join(base_opts)

    def configure_step2(self):
        """Configure the second stage of the bootstrap."""
        self._cmakeopts = {}
        self._configure_general_build()
        self._configure_intermediate_build()
        if self.full_llvm:
            self._cmakeopts.update(remove_gcc_dependency_opts)

    def configure_step3(self):
        """Configure the third stage of the bootstrap."""
        self._cmakeopts = {}
        self._configure_general_build()
        self._configure_final_build()
        # Update runtime CMake arguments, as they might have
        # changed when configuring the final build arguments
        self._add_cmake_runtime_args()
        if self.full_llvm:
            self._cmakeopts.update(remove_gcc_dependency_opts)

    @staticmethod
    def _create_compiler_config_file(compilers, gcc_prefix, installdir):
        """Create a config file for the compiler to point to the correct GCC installation."""
        bin_dir = os.path.join(installdir, 'bin')
        prefix_str = '--gcc-install-dir=%s' % gcc_prefix
        for comp in compilers:
            write_file(os.path.join(bin_dir, f'{comp}.cfg'), prefix_str)

    def build_with_prev_stage(self, prev_dir, stage_dir):
        """Build LLVM using the previous stage."""
        curdir = os.getcwd()

        bin_dir = os.path.join(prev_dir, 'bin')
        lib_dir_runtime = self.get_runtime_lib_path(prev_dir)

        # Give priority to the libraries in the current stage if compiled to avoid failures due to undefined symbols
        # e.g. when calling the compiled clang-ast-dump for stage 3
        lib_path = ':'.join(filter(None, [
            os.path.join(stage_dir, lib_dir_runtime),
            os.path.join(prev_dir, lib_dir_runtime),
        ]))

        if build_option('rpath'):
            self._prepare_runtimes_rpath_wrappers(stage_dir)

        # Needed for passing the variables to the build command
        with _wrap_env(bin_dir, lib_path):
            # If building with rpath, create RPATH wrappers for the Clang compilers for stage 2 and 3
            if build_option('rpath'):
                # !!! Should be replaced with ClangFlang (or correct naming) toolchain once available
                #     as this will only create rpath wrappers for Clang and not Flang
                my_toolchain = Clang(name='Clang', version='1')
                my_toolchain.prepare_rpath_wrappers(
                    rpath_include_dirs=[
                        os.path.join(self.installdir, 'lib'),
                        os.path.join(self.installdir, 'lib64'),
                        os.path.join(self.installdir, lib_dir_runtime),
                        ]
                    )
                self.log.info("Prepared rpath wrappers")

                # add symlink for 'opt' to wrapper dir, since Clang expects it in the same directory
                # see https://github.com/easybuilders/easybuild-easyblocks/issues/3075
                clang_wrapper_dir = os.path.dirname(which('clang'))
                symlink(os.path.join(prev_dir, 'opt'), os.path.join(clang_wrapper_dir, 'opt'))

                # RPATH wrappers add -Wl,rpath arguments to all command lines, including when it is just compiling
                # Clang by default warns about that, and then some configure tests use -Werror which turns those
                # warnings into errors. As a result, those configure tests fail, even though the compiler supports the
                # requested functionality (e.g. the test that checks if -fPIC is supported would fail, and it compiles
                # without resulting in relocation errors).
                # See https://github.com/easybuilders/easybuild-easyblocks/pull/2799#issuecomment-1270621100
                # Here, we add -Wno-unused-command-line-argument to CXXFLAGS to avoid these warnings alltogether
                cflags = os.getenv('CFLAGS', '')
                cxxflags = os.getenv('CXXFLAGS', '')
                setvar('CFLAGS', "%s %s" % (cflags, '-Wno-unused-command-line-argument'))
                setvar('CXXFLAGS', "%s %s" % (cxxflags, '-Wno-unused-command-line-argument'))

            if self.full_llvm:
                # See  https://github.com/llvm/llvm-project/issues/111667
                to_add = '--unwindlib=none'
                # for flags in ['CMAKE_C_FLAGS', 'CMAKE_CXX_FLAGS']:
                for flags in ['CMAKE_EXE_LINKER_FLAGS']:
                    ptr = self.runtimes_cmake_args[flags]
                    if to_add not in ptr:
                        ptr.append(to_add)

                self._add_cmake_runtime_args()

            # determine full path to clang/clang++ (which may be wrapper scripts in case of RPATH linking)
            clang = which('clang')
            clangxx = which('clang++')

            self._cmakeopts['CMAKE_C_COMPILER'] = clang
            self._cmakeopts['CMAKE_CXX_COMPILER'] = clangxx
            self._cmakeopts['CMAKE_ASM_COMPILER'] = clang
            self._cmakeopts['CMAKE_ASM_COMPILER_ID'] = 'Clang'

            # Also runs of the intermediate step compilers should be made aware of the GCC installation
            if LooseVersion(self.version) >= LooseVersion('19'):
                self._set_gcc_prefix()
                self._create_compiler_config_file(self.cfg_compilers, self.gcc_prefix, prev_dir)

            self.add_cmake_opts()

            change_dir(stage_dir)
            self.log.debug("Configuring %s", stage_dir)
            cmd = ' '.join(['cmake', self.cfg['configopts'], os.path.join(self.llvm_src_dir, 'llvm')])
            run_shell_cmd(cmd)

            self.log.debug("Building %s", stage_dir)
            cmd = f"make {self.make_parallel_opts} VERBOSE=1"
            res = run_shell_cmd(cmd, fail_on_error=False)
            # Observed in 20.1.0, the build of the offloading tools can fail due to 'cstdint' file not found
            # But will succeed if executed again with -j 1 (possible missing dependency in the CMake logic?)
            # See https://github.com/llvm/llvm-project/issues/130783
            if res.exit_code != EasyBuildExit.SUCCESS:
                self.log.warning("Build failed, attempting again with parallel ON")
                res = run_shell_cmd(cmd, fail_on_error=False)
            if res.exit_code != EasyBuildExit.SUCCESS:
                self.log.warning("Build failed, attempting again with parallel OFF")
                cmd = "make -j 1 VERBOSE=1"
                res = run_shell_cmd(cmd)

        change_dir(curdir)

    def _prepare_runtimes_rpath_wrappers(self, stage_dir):
        """Run the build command also ensuring proper rpath-ing for the Runtime build.

        Binaries and libraries produced during runtimes make use of the newly built Clang compiler which is not
        rpath-wrapped. This causes the executable to be produced without rpath (if required) and with
        runpath set to $ORIGIN. This causes 2 problems:
         - Binaries produced for the runtimes will fail the sanity check
         - Runtimes libraries that link to libLLVM.so like 'libomptarget.so' need LD_LIBRARY_PATH to work.
           This is because even if an executable compiled with the new llvm has rpath pointing to $EBROOTLLVM/lib,
           it will not be resolved with the executable's rpath, but the library's runpath
           (rpath is ignored if runpath is set).
           Even if libLLVM.so is a direct dependency of the executable, it needs to be resolved both for the
           executable and the library.

        Here we create a mock binary for the current stage, rpath-wrap it and than pass the rpath-wrapped binary
        to the runtimes build as the compiler through RUNTIMES_CMAKE_ARGS.
        """
        lib_dir_runtime = self.get_runtime_lib_path(stage_dir, fail_ok=True)

        # Give priority to the libraries in the current stage if compiled to avoid failures due to undefined symbols
        # e.g. when calling the compiled clang-ast-dump for stage 3
        lib_path = os.path.join(stage_dir, lib_dir_runtime)

        bin_dir_new = os.path.join(stage_dir, 'bin')
        mkdir(bin_dir_new, parents=True)
        with _wrap_env(bin_dir_new, lib_path):
            nxt_clang = os.path.join(bin_dir_new, 'clang')
            nxt_clangxx = os.path.join(bin_dir_new, 'clang++')
            with open(nxt_clang, 'w', encoding='utf-8') as f:
                f.write("#!/bin/bash\n")
                f.write("echo 'MOCK clang'\n")
            adjust_permissions(nxt_clang, stat.S_IXUSR)
            with open(nxt_clangxx, 'w', encoding='utf-8') as f:
                f.write("#!/bin/bash\n")
                f.write("echo 'MOCK clangxx'\n")
            adjust_permissions(nxt_clangxx, stat.S_IXUSR)

            tmp_toolchain = Clang(name='Clang', version='1')
            # Don't need stage dir here as LD_LIBRARY_PATH is set during build, this is only needed for
            # installed binaries with rpath
            lib_dirs = [os.path.join(self.installdir, x) for x in SEARCH_PATH_LIB_DIRS + [lib_dir_runtime]]
            tmp_toolchain.prepare_rpath_wrappers(rpath_include_dirs=lib_dirs)
            remove_file(nxt_clang)
            remove_file(nxt_clangxx)
            msg = "Prepared MOCK rpath wrappers needed to rpath-wrap also the new compilers produced "
            msg += "by the project build and than used for the runtimes build"
            self.log.info(msg)
            clang_mock = which('clang')
            clangxx_mock = which('clang++')

            clang_mock_wrapper_dir = os.path.dirname(clang_mock)

            symlink(os.path.join(stage_dir, 'opt'), os.path.join(clang_mock_wrapper_dir, 'opt'))

            # Use mocked rpath wrappers
            self.runtimes_cmake_args['CMAKE_C_COMPILER'] = [clang_mock]
            self.runtimes_cmake_args['CMAKE_CXX_COMPILER'] = [clangxx_mock]

        self._add_cmake_runtime_args()

    def build_step(self, *args, **kwargs):
        """Build LLVM, and optionally build it using itself."""
        if self.cfg['bootstrap']:
            self.log.info("Building stage 1")
            print_msg("Building stage 1/3")
        else:
            self.log.info("Building LLVM")
            print_msg("Building stage 1/1")

        change_dir(self.llvm_obj_dir_stage1)
        super().build_step(*args, **kwargs)

        if self.cfg['bootstrap']:
            self.log.info("Building stage 2")
            print_msg("Building stage 2/3")
            self.configure_step2()
            self.build_with_prev_stage(self.llvm_obj_dir_stage1, self.llvm_obj_dir_stage2)

            self.log.info("Building stage 3")
            print_msg("Building stage 3/3")
            self.configure_step3()
            self.build_with_prev_stage(self.llvm_obj_dir_stage2, self.llvm_obj_dir_stage3)

    def _para_test_step(self, parallel=1):
        """Run test suite with the specified number of parallel jobs for make."""
        basedir = self.final_dir

        # From grep -E "^[A-Z]+: " LOG_FILE | cut -d: -f1 | sort | uniq
        OUTCOMES_LOG = [
            'FAIL',
            'TIMEOUT',
        ]
        # OUTCOMES_OK = [
        #     'PASS',
        #     'UNSUPPORTED',
        #     'XFAIL',
        # ]

        change_dir(basedir)
        lib_path = ''
        if self.cfg['build_runtimes']:
            lib_dir_runtime = self.get_runtime_lib_path(basedir)
            lib_path = os.path.join(basedir, lib_dir_runtime)

        with _wrap_env(os.path.join(basedir, 'bin'), lib_path):
            cmd = f"make -j {parallel} check-all"
            res = run_shell_cmd(cmd, fail_on_error=False)
            out = res.output
            self.log.debug(out)

        ignore_patterns = self.ignore_patterns
        ignored_pattern_matches = 0
        failed_pattern_matches = 0
        if ignore_patterns:
            for line in out.splitlines():
                if any(line.startswith(f'{x}: ') for x in OUTCOMES_LOG):
                    if any(patt in line for patt in ignore_patterns):
                        self.log.info("Ignoring test failure: %s", line)
                        ignored_pattern_matches += 1
                    failed_pattern_matches += 1

        rgx_failed = re.compile(r'^ +Failed +: +([0-9]+)', flags=re.MULTILINE)
        mch = rgx_failed.search(out)
        if mch is None:
            rgx_passed = re.compile(r'^ +Passed +: +([0-9]+)', flags=re.MULTILINE)
            mch = rgx_passed.search(out)
            if mch is None:
                self.log.warning("Failed to extract number of failed/passed test results from output")
                num_failed = None
            else:
                num_failed = 0
        else:
            num_failed = int(mch.group(1))

        if num_failed is not None:
            num_timed_out = 0
            rgx_timed_out = re.compile(r'^ +Timed Out +: +([0-9]+)', flags=re.MULTILINE)
            mch = rgx_timed_out.search(out)
            if mch is not None:
                num_timed_out = int(mch.group(1))
                self.log.info("Tests timed out: %s", num_timed_out)
            num_failed += num_timed_out

        if num_failed != failed_pattern_matches:
            msg = f"Number of failed tests ({num_failed}) does not match "
            msg += f"number identified va line-by-line pattern matching: {failed_pattern_matches}"
            self.log.warning(msg)

        if num_failed is not None and ignored_pattern_matches:
            self.log.info("Ignored %s failed tests due to ignore patterns", ignored_pattern_matches)
            num_failed -= ignored_pattern_matches

        return num_failed

    def test_step(self):
        """Run tests on final stage (unless disabled)."""
        if not self.cfg['skip_all_tests']:
            # Also runs of test suite compilers should be made aware of the GCC installation
            if LooseVersion(self.version) >= LooseVersion('19'):
                self._set_gcc_prefix()
                self._create_compiler_config_file(self.cfg_compilers, self.gcc_prefix, self.final_dir)

            self.ignore_patterns = self.cfg['test_suite_ignore_patterns'] or []

            # For nvptx64 tests, find out if 'ptxas' exists in $PATH. If not, ignore all nvptx64 test failures
            pxtas_path = which('ptxas', on_error=IGNORE)
            if self.nvptx_target_cond and not pxtas_path:
                self.ignore_patterns += ['nvptx64-nvidia-cuda', 'nvptx64-nvidia-cuda-LTO']
                self.log.warning("PTXAS not found in PATH, ignoring failing tests for NVPTX target")
            # If the AMDGPU target is built, tests will be run if libhsa-runtime64.so is found.
            # However, this can cause issues if the system libraries are used, due to other loaded modules.
            # Therefore, ignore failing tests if ROCr-Runtime is not in the dependencies and
            # warn about this in the logs.
            if self.amdgpu_target_cond and 'rocr-runtime' not in self.deps:
                self.ignore_patterns += ['amdgcn-amd-amdhsa']
                self.log.warning("ROCr-Runtime not in dependencies, ignoring failing tests for AMDGPU target.")

            max_failed = self.cfg['test_suite_max_failed']
            num_failed = self._para_test_step(parallel=1)
            if num_failed is None:
                self.report_test_failure("Failed to extract test results from output")
                return

            if num_failed > max_failed:
                self.report_test_failure(f"Too many failed tests: {num_failed} ({max_failed} allowed)")
            elif num_failed:
                self.log.info(f"Test suite completed with {num_failed} failed tests ({max_failed} allowed)")
            else:
                self.log.info(f"Test suite completed, no failed tests ({max_failed} allowed)")

    def install_step(self):
        """Install stage 1 or 3 (if bootstrap) binaries."""
        basedir = self.final_dir
        change_dir(basedir)

        if self.cfg['build_runtimes']:
            orig_ld_library_path = os.getenv('LD_LIBRARY_PATH')
            lib_dir_runtime = self.get_runtime_lib_path(basedir)

            lib_path = ':'.join([
                os.path.join(basedir, lib_dir_runtime),
                orig_ld_library_path
            ])

            self.cfg.update('preinstallopts', f'LD_LIBRARY_PATH={lib_path}')

        super().install_step()

        # copy Python bindings here in post-install step so that it is not done more than once in multi_deps context
        if self.cfg['python_bindings']:
            python_bindings_source_dir = os.path.join(self.llvm_src_dir, 'clang', 'bindings', 'python')
            python_bindins_target_dir = os.path.join(self.installdir, 'lib', 'python')
            copy_dir(python_bindings_source_dir, python_bindins_target_dir, dirs_exist_ok=True)

            python_bindings_source_dir = os.path.join(self.llvm_src_dir, 'mlir', 'python')
            copy_dir(python_bindings_source_dir, python_bindins_target_dir, dirs_exist_ok=True)

        if LooseVersion(self.version) >= LooseVersion('19'):
            # For GCC aware installation create config files in order to point to the correct GCC installation
            # Required as GCC_INSTALL_PREFIX was removed (see https://github.com/llvm/llvm-project/pull/87360)
            self._set_gcc_prefix()
            self._create_compiler_config_file(self.cfg_compilers, self.gcc_prefix, self.installdir)

        # This is needed as some older build system will select a different naming scheme for the library leading to
        # The correct target <__config_site> and libclang_rt.builtins.a not being found
        # An example is building BOOST
        resdir_version = self.version.split('.')[0]
        clang_lib = os.path.join(self.installdir, 'lib', 'clang', resdir_version, 'lib')

        for orig, other in self.symlink_lst:
            for dirname in ['include', 'lib', clang_lib]:
                src = os.path.join(self.installdir, dirname, orig)
                dst = os.path.join(self.installdir, dirname, other)
                if os.path.exists(src) and not os.path.exists(dst):
                    msg = f"Creating symlink for {src} to {dst} for better compatibility with expected --target"
                    self.log.info(msg)
                    symlink(src, dst)

    def get_runtime_lib_path(self, base_dir, fail_ok=False):
        """Return the path to the runtime libraries."""
        if self.host_triple is None:
            # Attempt using the glob based detection of the runtime library directory for runs of
            # --sanity-check-only/--module-only where the configure step is not used
            arch = get_arch_prefix()
            glob_pattern = os.path.join(base_dir, 'lib', f'{arch}-*')
            matches = glob.glob(glob_pattern)
            if matches:
                self.host_triple = os.path.basename(matches[0])
            else:
                raise EasyBuildError("Could not find runtime library directory")

        res = os.path.join('lib', self.host_triple)
        if not fail_ok:
            path = os.path.join(base_dir, res)
            if not os.path.exists(path):
                raise EasyBuildError("Could not find runtime library directory '%s'", path)

        return res

    def banned_linked_shared_libs(self):
        """Return a list of shared libraries that should not be linked against."""
        res = []
        if self.full_llvm:
            self.log.info("Checking that no GCC shared libraries are linked against")
            res += ['libstdc++', 'libgcc_s', 'libicuuc']
        if not self.build_shared:
            # Libraries should be linked statically
            self.log.info("Checking that no shared libraries are linked against in static build")
            res += ['libc++', 'libc++abi', 'libunwind']
        return res

    @staticmethod
    def _sanity_check_gcc_prefix(compilers, gcc_prefix, installdir):
        """Check if the GCC prefix of the compiler is correct"""
        rgx = re.compile('Selected GCC installation: (.*)')
        for comp in compilers:
            cmd = "%s -v" % os.path.join(installdir, 'bin', comp)
            res = run_shell_cmd(cmd, fail_on_error=False)
            out = res.output
            mch = rgx.search(out)
            if mch is None:
                raise EasyBuildError("Failed to extract GCC installation path from output of '%s': %s", cmd, out)
            check_prefix = mch.group(1)
            if check_prefix != gcc_prefix:
                error_msg = f"GCC installation path '{check_prefix}' does not match expected path '{gcc_prefix}'"
                raise EasyBuildError(error_msg)

    def sanity_check_step(self, custom_paths=None, custom_commands=None, extension=False, extra_modules=None):
        """Perform sanity checks on the installed LLVM."""
        lib_dir_runtime = None
        if self.cfg['build_runtimes']:
            lib_dir_runtime = self.get_runtime_lib_path(self.installdir)
        shlib_ext = '.' + get_shared_lib_ext()

        resdir_version = self.version.split('.')[0]

        # Detect OpenMP support for CPU architecture
        arch = get_cpu_architecture()
        # Check architecture explicitly since Clang uses potentially different names
        if arch == X86_64:
            arch = 'x86_64'
        elif arch == POWER:
            arch = 'ppc64'
        elif arch == AARCH64:
            arch = 'aarch64'

        check_files = []
        check_bin_files = []
        check_lib_files = []
        check_librt_files = []
        check_inc_files = []
        check_dirs = ['include/llvm', 'include/llvm-c', 'lib/cmake/llvm']
        custom_commands = [
            "llvm-ar --help",
            "llvm-ranlib --help",
            "llvm-nm --help",
            "llvm-objdump --help",
        ]
        gcc_prefix_compilers = []
        if self.build_shared:
            check_lib_files += ['libLLVM.so']

        if 'clang' in self.final_projects:
            check_bin_files += [
                'clang', 'clang++', 'clang-cpp', 'clang-cl', 'clang-repl', 'hmaptool', 'amdgpu-arch', 'nvptx-arch',
                'intercept-build', 'scan-build', 'scan-build-py', 'scan-view', 'analyze-build', 'c-index-test',
                'clang-tblgen',
            ]
            check_lib_files += [
                'libclang.so', 'libclang-cpp.so', 'libclangAST.a', 'libclangCrossTU.a', 'libclangFrontend.a',
                'libclangInterpreter.a', 'libclangParse.a', 'libclangTooling.a'
            ]
            check_dirs += [
                'lib/cmake/clang', 'include/clang'
            ]
            custom_commands += ['llvm-config --cxxflags', 'clang --help', 'clang++ --help']
            gcc_prefix_compilers += ['clang', 'clang++']

        if 'clang-tools-extra' in self.final_projects:
            # clang-pseudo removed with LLVM 20
            check_bin_files += [
                'clangd', 'clang-tidy', 'clang-include-fixer', 'clang-query', 'clang-move',
                'clang-reorder-fields', 'clang-include-cleaner', 'clang-apply-replacements',
                'clang-change-namespace', 'pp-trace', 'modularize'
            ]
            check_lib_files += [
                'libclangTidy.a', 'libclangQuery.a', 'libclangIncludeFixer.a', 'libclangIncludeCleaner.a',
            ]
            check_dirs += ['include/clang-tidy']
        if 'flang' in self.final_projects:
            if LooseVersion(self.version) < LooseVersion('19'):
                check_bin_files += ['bbc', 'flang-new', 'flang-to-external-fc', 'f18-parse-demo', 'fir-opt', 'tco']
            else:
                check_bin_files += ['bbc', 'flang-new', 'f18-parse-demo', 'fir-opt', 'tco']
            check_lib_files += [
                'libFortranRuntime.a', 'libFortranSemantics.a', 'libFortranLower.a', 'libFortranParser.a',
                'libFIRCodeGen.a', 'libflangFrontend.a', 'libFortranCommon.a', 'libFortranDecimal.a',
                'libHLFIRDialect.a'
            ]
            check_dirs += ['lib/cmake/flang', 'include/flang']
            custom_commands += ['bbc --help', 'mlir-tblgen --help', 'flang-new --help']
            gcc_prefix_compilers += ['flang-new']

        if 'lld' in self.final_projects:
            check_bin_files += ['ld.lld', 'lld', 'lld-link', 'wasm-ld']
            check_lib_files += [
                'liblldCOFF.a', 'liblldCommon.a', 'liblldELF.a', 'liblldMachO.a', 'liblldMinGW.a', 'liblldWasm.a'
            ]
            check_dirs += ['lib/cmake/lld', 'include/lld']
        if 'lldb' in self.final_projects:
            check_bin_files += ['lldb']
            if self.build_shared:
                check_lib_files += ['liblldb.so']
            check_dirs += ['include/lldb']
        if 'mlir' in self.final_projects:
            check_bin_files += ['mlir-opt', 'tblgen-to-irdl', 'mlir-pdll']
            check_lib_files += [
                'libMLIRIR.a', 'libmlir_async_runtime.so', 'libmlir_arm_runner_utils.so', 'libmlir_c_runner_utils.so',
                'libmlir_float16_utils.so'
            ]
            check_dirs += ['lib/cmake/mlir', 'include/mlir', 'include/mlir-c']
        if 'libunwind' in self.final_runtimes:
            check_librt_files += ['libunwind.a']
            if self.build_shared:
                check_librt_files += ['libunwind.so']
            check_inc_files += ['unwind.h', 'libunwind.h', 'mach-o/compact_unwind_encoding.h']
        if 'libcxx' in self.final_runtimes:
            check_librt_files += ['libc++.a']
            if self.build_shared:
                check_librt_files += ['libc++.so']
            check_dirs += ['include/c++/v1']
        if 'libcxxabi' in self.final_runtimes:
            check_librt_files += ['libc++abi.a']
            if self.build_shared:
                check_librt_files += ['libc++abi.so']

        if 'polly' in self.final_projects:
            check_lib_files += ['libPolly.a', 'libPollyISL.a']
            if self.build_shared:
                check_lib_files += ['LLVMPolly.so']
            check_dirs += ['lib/cmake/polly', 'include/polly']
            custom_commands += [
                ' | '.join([
                    'echo \'int main(int argc, char **argv) { return 0; }\'',
                    'clang -xc -O3 -mllvm -polly -'
                ]) + ' && ./a.out && rm -f a.out'
            ]
        if 'bolt' in self.final_projects:
            check_bin_files += ['llvm-bolt', 'llvm-boltdiff', 'llvm-bolt-heatmap']
            check_lib_files += ['libbolt_rt_instr.a']
            custom_commands += ['llvm-bolt --help']
        if 'openmp' in self.final_projects:
            omp_lib_files = []
            omp_lib_files += ['libomp.so', 'libompd.so']
            if self.cfg['build_openmp_offload']:
                # Judging from the build process/logs of LLVM 19, the omptarget plugins (rtl.<device>.so) are now built
                # as static libraries and linked into the libomptarget.so shared library
                omp_lib_files += ['libomptarget.so']
                if LooseVersion(self.version) < LooseVersion('19'):
                    omp_lib_files += ['libomptarget.rtl.%s.so' % arch]
                if self.nvptx_target_cond:
                    if LooseVersion(self.version) < LooseVersion('19'):
                        omp_lib_files += ['libomptarget.rtl.cuda.so']
                    if LooseVersion(self.version) < LooseVersion('20'):
                        omp_lib_files += ['libomptarget-nvptx-sm_%s.bc' % cc for cc in self.cuda_cc]
                    else:
                        omp_lib_files += ['libomptarget-nvptx.bc']
                if self.amdgpu_target_cond:
                    if LooseVersion(self.version) < LooseVersion('19'):
                        omp_lib_files += ['libomptarget.rtl.amdgpu.so']
                    if LooseVersion(self.version) < LooseVersion('20'):
                        omp_lib_files += ['libomptarget-amdgpu-%s.bc' % gfx for gfx in self.amd_gfx]
                    else:
                        omp_lib_files += ['libomptarget-amdgpu.bc']

                if LooseVersion(self.version) < LooseVersion('19'):
                    # Before LLVM 19, omp related libraries are installed under 'ROOT/lib''
                    check_lib_files += omp_lib_files
                else:
                    # Starting from LLVM 19, omp related libraries are installed the runtime library directory
                    check_librt_files += omp_lib_files
                    check_bin_files += ['llvm-omp-kernel-replay']
                    if LooseVersion(self.version) < LooseVersion('20'):
                        check_bin_files += ['llvm-omp-device-info']
                    else:
                        check_bin_files += ['llvm-offload-device-info']

        if self.cfg['build_openmp_tools']:
            check_files += [os.path.join('lib', 'clang', resdir_version, 'include', 'ompt.h')]
            if LooseVersion(self.version) < LooseVersion('19'):
                check_lib_files += ['libarcher.so']
            elif LooseVersion(self.version) >= LooseVersion('19'):
                check_librt_files += ['libarcher.so']
        if self.cfg['python_bindings']:
            custom_commands += ["python -c 'import clang'"]
            custom_commands += ["python -c 'import mlir'"]

        check_files += [os.path.join('bin', x) for x in check_bin_files]
        check_files += [os.path.join('lib', x) for x in check_lib_files]
        check_files += [os.path.join(lib_dir_runtime, x) for x in check_librt_files]
        check_files += [os.path.join('include', x) for x in check_inc_files]

        for libso in check_files:
            if not libso.endswith('.so'):
                continue
            libext = libso.replace('.so', shlib_ext)
            if libext not in check_files:
                check_files.remove(libso)
                check_files.append(libext)

        custom_paths = {
            'files': check_files,
            'dirs': check_dirs,
        }

        self._set_gcc_prefix()
        if lib_dir_runtime:
            # Required for 'clang -v' to work if linked to LLVM runtimes
            with _wrap_env(ld_path=os.path.join(self.installdir, lib_dir_runtime)):
                self._sanity_check_gcc_prefix(gcc_prefix_compilers, self.gcc_prefix, self.installdir)
        else:
            self._sanity_check_gcc_prefix(gcc_prefix_compilers, self.gcc_prefix, self.installdir)

        return super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_step(self, *args, **kwargs):
        """
        Clang can find its own headers and libraries but the shared libraries need to be in $LD_LIBRARY_PATH
        """
        mod_env_headers = self.module_load_environment.alias_vars(MODULE_LOAD_ENV_HEADERS)
        for disallowed_var in mod_env_headers:
            self.module_load_environment.remove(disallowed_var)
            self.log.debug(f"Purposely not updating ${disallowed_var} in {self.name} module file")

        lib_dirs = SEARCH_PATH_LIB_DIRS[:]
        if self.cfg['build_runtimes']:
            runtime_lib_path = self.get_runtime_lib_path(self.installdir)
            lib_dirs.append(runtime_lib_path)

        self.log.debug(f"List of subdirectories for libraries to add to $LD_LIBRARY_PATH + $LIBRARY_PATH: {lib_dirs}")
        self.module_load_environment.LD_LIBRARY_PATH = lib_dirs
        self.module_load_environment.LIBRARY_PATH = lib_dirs

        return super().make_module_step(*args, **kwargs)

    def make_module_extra(self):
        """Custom variables for Clang module."""
        txt = super().make_module_extra()
        # we set the symbolizer path so that asan/tsan give meanfull output by default
        asan_symbolizer_path = os.path.join(self.installdir, 'bin', 'llvm-symbolizer')
        txt += self.module_generator.set_environment('ASAN_SYMBOLIZER_PATH', asan_symbolizer_path)
        if self.cfg['python_bindings']:
            txt += self.module_generator.prepend_paths('PYTHONPATH', os.path.join('lib', 'python'))
        return txt



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing ScaLAPACK, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import glob
import os
from easybuild.tools import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.toolchains.linalg.acml import Acml
from easybuild.toolchains.linalg.atlas import Atlas
from easybuild.toolchains.linalg.blis import Blis
from easybuild.toolchains.linalg.flexiblas import FlexiBLAS, det_flexiblas_backend_libs
from easybuild.toolchains.linalg.gotoblas import GotoBLAS
from easybuild.toolchains.linalg.lapack import Lapack
from easybuild.toolchains.linalg.openblas import OpenBLAS
from easybuild.toolchains.linalg.intelmkl import IntelMKL
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file, remove_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_shell_cmd


class EB_ScaLAPACK(CMakeMake):
    """
    Support for building and installing ScaLAPACK, both versions 1.x and 2.x
    """

    def __init__(self, *args, **kwargs):
        """Constructor of ScaLAPACK easyblock."""
        super().__init__(*args, **kwargs)

        self.loosever = LooseVersion(self.version)

        # use CMake for recent versions, but only if CMake is listed as a build dep
        build_deps_names = [dep['name'].lower() for dep in self.cfg.builddependencies()]
        self.use_cmake = self.loosever >= LooseVersion('2.1.0') and 'cmake' in build_deps_names

    def configure_step(self):
        """Configure ScaLAPACK build by copying SLmake.inc.example to SLmake.inc and checking dependencies."""

        # use CMake for recent versions, but only if CMake is listed as a build dep
        if self.use_cmake:
            super().configure_step()
        else:
            src = os.path.join(self.cfg['start_dir'], 'SLmake.inc.example')
            dest = os.path.join(self.cfg['start_dir'], 'SLmake.inc')

            if os.path.exists(dest):
                raise EasyBuildError("Destination file %s exists", dest)
            else:
                copy_file(src, dest)

    def build_libscalapack_make(self):
        """Build libscalapack using 'make -j', after determining the options to pass to make."""
        # MPI compiler commands
        known_mpi_libs = [toolchain.MPICH, toolchain.MPICH2, toolchain.MVAPICH2]  # @UndefinedVariable
        known_mpi_libs += [toolchain.OPENMPI, toolchain.QLOGICMPI]  # @UndefinedVariable
        known_mpi_libs += [toolchain.INTELMPI]  # @UndefinedVariable
        if os.getenv('MPICC') and os.getenv('MPIF90'):
            mpicc = os.getenv('MPICC')
            mpif90 = os.getenv('MPIF90')
        elif self.toolchain.mpi_family() in known_mpi_libs:
            mpicc = 'mpicc'
            mpif90 = 'mpif90'
        else:
            raise EasyBuildError("Don't know which compiler commands to use.")

        # determine build options BLAS and LAPACK libs
        extra_makeopts = []

        acml = get_software_root(Acml.LAPACK_MODULE_NAME[0])
        flexiblas = get_software_root(FlexiBLAS.LAPACK_MODULE_NAME[0])
        intelmkl = get_software_root(IntelMKL.LAPACK_MODULE_NAME[0])
        lapack = get_software_root(Lapack.LAPACK_MODULE_NAME[0])
        openblas = get_software_root(OpenBLAS.LAPACK_MODULE_NAME[0])

        if flexiblas:
            libdir = os.path.join(flexiblas, 'lib')
            blas_libs = ' '.join(['-l%s' % lib for lib in FlexiBLAS.BLAS_LIB])
            extra_makeopts.extend([
                'BLASLIB="-L%s %s -lpthread"' % (libdir, blas_libs),
                'LAPACKLIB="-L%s %s"' % (libdir, blas_libs),
            ])
        elif lapack:
            extra_makeopts.append('LAPACKLIB=%s' % os.path.join(lapack, 'lib', 'liblapack.a'))

            for blas in [Atlas, Blis, GotoBLAS]:
                blas_root = get_software_root(blas.BLAS_MODULE_NAME[0])
                if blas_root:
                    blas_libs = ' '.join(['-l%s' % lib for lib in blas.BLAS_LIB])
                    blas_libdir = os.path.join(blas_root, 'lib')
                    extra_makeopts.append('BLASLIB="-L%s %s -lpthread"' % (blas_libdir, blas_libs))
                    break

            if not blas_root:
                raise EasyBuildError("Failed to find a known BLAS library, don't know how to define 'BLASLIB'")

        elif acml:
            acml_base_dir = os.getenv('ACML_BASEDIR', 'NO_ACML_BASEDIR')
            acml_static_lib = os.path.join(acml, acml_base_dir, 'lib', 'libacml.a')
            extra_makeopts.extend([
                'BLASLIB="%s -lpthread"' % acml_static_lib,
                'LAPACKLIB=%s' % acml_static_lib
            ])
        elif openblas:
            libdir = os.path.join(openblas, 'lib')
            blas_libs = ' '.join(['-l%s' % lib for lib in OpenBLAS.BLAS_LIB])
            extra_makeopts.extend([
                'BLASLIB="-L%s %s -lpthread"' % (libdir, blas_libs),
                'LAPACKLIB="-L%s %s"' % (libdir, blas_libs),
            ])
        elif intelmkl:
            libdir = os.path.join(intelmkl, 'mkl', 'lib', 'intel64')
            blas_libs = os.environ['LIBLAPACK']
            extra_makeopts.extend([
                'BLASLIB="-L%s %s -lpthread"' % (libdir, blas_libs),
                'LAPACKLIB="-L%s %s"' % (libdir, blas_libs),
            ])
        else:
            raise EasyBuildError("Unknown LAPACK library used, no idea how to set BLASLIB/LAPACKLIB make options")

        # determine interface
        if self.toolchain.mpi_family() in known_mpi_libs:
            interface = 'Add_'
        else:
            raise EasyBuildError("Don't know which interface to pick for the MPI library being used.")

        # set compilers and options
        extra_makeopts += [
            'FC="%s"' % mpif90,
            'CC="%s"' % mpicc,
            'CCFLAGS="%s"' % os.getenv('CFLAGS'),
            'FCFLAGS="%s"' % os.getenv('FFLAGS'),
        ]

        # set interface
        extra_makeopts.append('CDEFS="-D%s"' % interface)

        # update make opts, and build_step
        saved_buildopts = self.cfg['buildopts']

        # Only build the library first, that can be done in parallel.
        # Creating libscalapack.a may fail in parallel, but should work
        # fine with non-parallel make afterwards
        self.cfg.update('buildopts', 'lib')
        self.cfg.update('buildopts', ' '.join(extra_makeopts))

        cmd = "%s make %s %s" % (self.cfg['prebuildopts'], self.parallel_flag, self.cfg['buildopts'])

        # Ignore exit code for parallel run
        run_shell_cmd(cmd, fail_on_error=False)

        # Now prepare to remake libscalapack.a serially and the tests.
        self.cfg['buildopts'] = saved_buildopts
        self.cfg.update('buildopts', ' '.join(extra_makeopts))

        remove_file('libscalapack.a')
        self.cfg.parallel = 1

    def build_step(self):
        """Build ScaLAPACK using make after setting make options."""

        # only do a parallel pre-build of libscalapack and set up build options if we're not using CMake
        if not self.use_cmake:
            self.build_libscalapack_make()

        super().build_step()

    def install_step(self):
        """Install by copying files to install dir."""

        if self.use_cmake:
            super().install_step()
        else:
            # 'manually' install ScaLAPACK by copying headers and libraries if we're not using CMake
            path_info = [
                ('SRC', 'include', '.h'),  # include files
                ('', 'lib', '.a'),  # libraries
            ]
            for (srcdir, destdir, ext) in path_info:

                src = os.path.join(self.cfg['start_dir'], srcdir)
                dest = os.path.join(self.installdir, destdir)

                for lib in glob.glob(os.path.join(src, '*%s' % ext)):
                    copy_file(lib, os.path.join(dest, os.path.basename(lib)))
                    self.log.debug("Copied %s to %s", lib, dest)

    def banned_linked_shared_libs(self):
        """
        List of shared libraries which are not allowed to be linked in any installed binary/library.
        """
        res = super().banned_linked_shared_libs()

        # register FlexiBLAS backends as banned libraries,
        # ScaLAPACK should not be linking to those directly
        if get_software_root(FlexiBLAS.LAPACK_MODULE_NAME[0]):
            res.extend(det_flexiblas_backend_libs())

        return res

    def sanity_check_step(self):
        """Custom sanity check for ScaLAPACK."""

        custom_paths = {
            'files': [os.path.join('lib', 'libscalapack.a')],
            'dirs': []
        }

        super().sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2015-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for using (already installed/existing) system compiler based
on GCC instead of a full install via EasyBuild.

@author Bernd Mohr (Juelich Supercomputing Centre)
@author Kenneth Hoste (Ghent University)
@author Alan O'Cais (Juelich Supercomputing Centre)
@author Alex Domingo (Vrije Universiteit Brussel)
"""
from easybuild.easyblocks.gcc import EB_GCC
from easybuild.easyblocks.generic.systemcompiler import SystemCompiler


# order matters, SystemCompiler goes first to avoid recursion whenever EB_GCC calls super()
class SystemCompilerGCC(SystemCompiler, EB_GCC):
    """
    Support for generating a module file for a system compiler based on GCC with specified name.

    The compiler is expected to be available in $PATH, required libraries are assumed to be readily available.

    Specifying 'system' as a version leads to using the derived compiler version in the generated module;
    if an actual version is specified, it is checked against the derived version of the system compiler that was found.
    """
    @staticmethod
    def extra_options():
        """Add custom easyconfig parameters for SystemCompilerGCC easyblock."""
        extra_vars = EB_GCC.extra_options()
        extra_vars.update(SystemCompiler.extra_options())
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Extra initialization: keep track of values that may change due to modifications to the version."""
        super().__init__(*args, **kwargs)

        # use GCC compiler class to generate standalone module
        if self.cfg['generate_standalone_module']:
            self.compiler_class = EB_GCC



##
# Copyright 2020-2025 Forschungszentrum Juelich GmbH
#
# This file is triple-licensed under GPLv2 (see below), MIT, and
# BSD three-clause licenses.
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Support for installing AOCC, implemented as an easyblock.

@author: Sebastian Achilles (Forschungszentrum Juelich GmbH)
"""

import glob
import os
import re
import stat
import tempfile

from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, move_file, write_file
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext, get_cpu_architecture

# Wrapper script definition
WRAPPER_TEMPLATE = """#!/bin/bash

# Patch argv[0] to the actual compiler so that the correct driver is used internally
(exec -a "$0" {compiler_name} --gcc-toolchain=$EBROOTGCCCORE "$@")
"""

AOCC_MINIMAL_CPP_EXAMPLE = """
#include <iostream>

int main(){ std::cout << "It works!" << std::endl; }
"""

AOCC_MINIMAL_FORTRAN_EXAMPLE = """
program main
end program main
"""


class EB_AOCC(PackedBinary):
    """
    Support for installing the AOCC compilers
    """

    @staticmethod
    def extra_options():
        extra_vars = {
            'clangversion': [None, "Clang Version on which AOCC is based on (10.0.0 or 11.0.0 or ...)", CUSTOM],
        }
        return PackedBinary.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Easyblock constructor, define custom class variables specific to AOCC."""
        super().__init__(*args, **kwargs)

        self.clangversion = self.cfg['clangversion']
        # AOCC is based on Clang. Try to guess the clangversion from the AOCC version
        # if clangversion is not specified in the easyconfig
        if self.clangversion is None:
            self.clangversion = self._aocc_guess_clang_version()

        self._gcc_prefix = None

        # Bypass the .mod file check for GCCcore installs
        self.cfg['skip_mod_files_sanity_check'] = True

    def _aocc_guess_clang_version(self):
        map_aocc_to_clang_ver = {
            '2.3.0': '11.0.0',
            '3.0.0': '12.0.0',
            '3.1.0': '12.0.0',
            '3.2.0': '13.0.0',
            '4.0.0': '14.0.6',
            '4.1.0': '16.0.3',
            '4.2.0': '16.0.3',
            '5.0.0': '17.0.6',
        }

        if self.version in map_aocc_to_clang_ver:
            return map_aocc_to_clang_ver[self.version]
        else:
            error_lines = [
                "AOCC is based on Clang. Guessing Clang version in easyblock failed.",
                "You should either:",
                "- specify `clangversion` in the easyconfig;",
                "- extend `map_aocc_to_clang_ver` in the easyblock;",
            ]
            raise EasyBuildError('\n'.join(error_lines))

    def _create_compiler_wrappers(self, compilers_to_wrap):
        if not compilers_to_wrap:
            return

        orig_compiler_tmpl = f"{os.path.join(self.installdir, 'bin')}/{{}}.orig"

        def create_wrapper(wrapper_comp):
            """Create for a particular compiler, with a particular name"""
            wrapper_f = os.path.join(self.installdir, 'bin', wrapper_comp)
            compiler_name = orig_compiler_tmpl.format(wrapper_comp)
            write_file(wrapper_f, WRAPPER_TEMPLATE.format(compiler_name=compiler_name))

            perms = stat.S_IXUSR | stat.S_IRUSR | stat.S_IXGRP | stat.S_IRGRP | stat.S_IXOTH | stat.S_IROTH
            adjust_permissions(wrapper_f, perms)

        # Rename original compilers and prepare wrappers to pick up GCCcore as GCC toolchain for the compilers
        for comp in compilers_to_wrap:
            actual_compiler = os.path.join(self.installdir, 'bin', comp)
            if os.path.isfile(actual_compiler):
                move_file(actual_compiler, orig_compiler_tmpl.format(comp))
            else:
                raise EasyBuildError(f"Cannot make wrapper for '{actual_compiler}', file does not exist")

            if not os.path.exists(actual_compiler):
                create_wrapper(comp)
                self.log.info(f"Wrapper for {comp} successfully created")
            else:
                raise EasyBuildError(f"Cannot make wrapper for '{actual_compiler}', original compiler was not renamed!")

    def _create_compiler_config_files(self, compilers_to_add_config_file):
        """For each of the compiler suites, add a .cfg file which points to the correct GCCcore as the GCC toolchain."""
        if not compilers_to_add_config_file:
            return

        bin_dir = os.path.join(self.installdir, 'bin')
        prefix_str = '--gcc-install-dir=%s' % self.gcc_prefix
        for comp in compilers_to_add_config_file:
            with open(os.path.join(bin_dir, '%s.cfg' % comp), 'w') as f:
                f.write(prefix_str)

    def _sanity_check_gcc_prefix(self):
        """Check if the GCC prefix is correct."""
        compilers_to_check = [
            'clang',
            'clang++',
            'clang-%s' % LooseVersion(self.clangversion).version[0],
            'clang-cpp',
            'flang',
        ]

        rgx = re.compile('Selected GCC installation: (.*)')
        for comp in compilers_to_check:
            cmd = "%s -v" % os.path.join(self.installdir, 'bin', comp)
            res = run_shell_cmd(cmd, fail_on_error=False)
            mch = rgx.search(res.output)
            if mch is None:
                self.log.debug(res.output)
                raise EasyBuildError("Failed to extract GCC installation path from output of `%s`", cmd)
            gcc_prefix = mch.group(1)
            if gcc_prefix != self.gcc_prefix:
                raise EasyBuildError(
                    "GCC installation path `%s` does not match expected path `%s`", gcc_prefix, self.gcc_prefix
                    )

    @property
    def gcc_prefix(self):
        """Set the GCC prefix for the build."""
        if not self._gcc_prefix:
            arch = get_cpu_architecture()
            gcc_root = get_software_root('GCCcore')
            gcc_version = get_software_version('GCCcore')
            # If that doesn't work, try with GCC
            if gcc_root is None:
                gcc_root = get_software_root('GCC')
                gcc_version = get_software_version('GCC')
            # If that doesn't work either, print error and exit
            if gcc_root is None:
                raise EasyBuildError("Can't find GCC or GCCcore to use")

            pattern = os.path.join(gcc_root, 'lib', 'gcc', '%s-*' % arch, '%s' % gcc_version)
            matches = glob.glob(pattern)
            if not matches:
                raise EasyBuildError("Can't find GCC version %s for architecture %s in %s", gcc_version, arch, pattern)
            self._gcc_prefix = os.path.abspath(matches[0])
            self.log.debug("Using %s as the gcc install location", self._gcc_prefix)

        return self._gcc_prefix

    def install_step(self):
        # EULA for AOCC must be accepted via --accept-eula-for EasyBuild configuration option,
        # or via 'accept_eula = True' in easyconfig file
        self.check_accepted_eula(more_info='http://developer.amd.com/wordpress/media/files/AOCC_EULA.pdf')

        super().install_step()

    def post_processing_step(self):
        """
        For AOCC <5.0.0:
        Create wrappers for the compilers to make sure compilers picks up GCCcore as GCC toolchain.
        For AOCC >= 5.0.0:
        Create [compiler-name].cfg files to point the compiler to the correct GCCcore as GCC toolchain.
        For compilers not supporting this option, wrap the compilers using the old method.
        """
        compilers_to_wrap = []
        compilers_to_add_config_files = []

        if LooseVersion(self.version) < LooseVersion("5.0.0"):
            compilers_to_wrap += [
                f'clang-{LooseVersion(self.clangversion).version[0]}',
            ]
            if not self.cfg['keepsymlinks']:
                compilers_to_wrap += [
                    'clang',
                    'clang++',
                    'clang-cpp',
                    'flang',
                ]
        else:
            compilers_to_add_config_files += [
                'clang',
                'clang++',
                'clang-cpp'
            ]
            compilers_to_wrap += [
                'flang'
            ]

        self._create_compiler_config_files(compilers_to_add_config_files)
        self._create_compiler_wrappers(compilers_to_wrap)
        super().post_processing_step()

    def sanity_check_step(self):
        """Custom sanity check for AOCC, based on sanity check for Clang."""

        # Clang v16+ only use the major version number for the resource dir
        resdir_version = self.clangversion
        if LooseVersion(self.clangversion) >= LooseVersion('16.0.0'):
            resdir_version = LooseVersion(self.clangversion).version[0]

        shlib_ext = get_shared_lib_ext()
        custom_paths = {
            'files': [
                'bin/clang', 'bin/clang++', 'bin/flang', 'bin/lld', 'bin/llvm-ar', 'bin/llvm-as', 'bin/llvm-config',
                'bin/llvm-link', 'bin/llvm-nm', 'bin/llvm-symbolizer', 'bin/opt', 'bin/scan-build', 'bin/scan-view',
                'include/clang-c/Index.h', 'include/llvm-c/Core.h', 'lib/clang/%s/include/omp.h' % resdir_version,
                'lib/clang/%s/include/stddef.h' % resdir_version, 'lib/libclang.%s' % shlib_ext,
                'lib/libomp.%s' % shlib_ext,
            ],
            'dirs': ['include/llvm', 'lib/clang/%s/lib' % resdir_version, 'lib32'],
        }

        custom_commands = [
            "clang --help",
            "clang++ --help",
            "clang-%s --help" % LooseVersion(self.clangversion).version[0],
            "clang-cpp --help",
            "flang --help",
            "llvm-config --cxxflags",
        ]

        self._sanity_check_gcc_prefix()

        # Check if clang++ can actually compile programs. This can fail if the wrong driver is picked up by LLVM.
        tmpdir = tempfile.mkdtemp()
        write_file(os.path.join(tmpdir, 'minimal.cpp'), AOCC_MINIMAL_CPP_EXAMPLE)
        minimal_cpp_compiler_cmd = "cd %s && clang++ minimal.cpp -o minimal_cpp" % tmpdir
        custom_commands.append(minimal_cpp_compiler_cmd)
        # Check if flang can actually compile programs. This can fail if the wrong driver is picked up by LLVM.
        write_file(os.path.join(tmpdir, 'minimal.f90'), AOCC_MINIMAL_FORTRAN_EXAMPLE)
        minimal_f90_compiler_cmd = "cd %s && flang minimal.f90 -o minimal_f90" % tmpdir
        custom_commands.append(minimal_f90_compiler_cmd)

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_extra(self):
        """Custom variables for AOCC module."""
        txt = super().make_module_extra()
        # we set the symbolizer path so that asan/tsan give meanfull output by default
        asan_symbolizer_path = os.path.join(self.installdir, 'bin', 'llvm-symbolizer')
        txt += self.module_generator.set_environment('ASAN_SYMBOLIZER_PATH', asan_symbolizer_path)
        # setting the AOCChome path
        txt += self.module_generator.set_environment('AOCChome', self.installdir)
        return txt



##
# Copyright 2015-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for using (already installed/existing) system compiler
instead of a full install via EasyBuild.

@author Bernd Mohr (Juelich Supercomputing Centre)
@author Kenneth Hoste (Ghent University)
@author Alan O'Cais (Juelich Supercomputing Centre)
@author Alex Domingo (Vrije Universiteit Brussel)
"""
import os
import re

from easybuild.base import fancylogger
from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.filetools import read_file, resolve_path, which
from easybuild.tools.run import run_shell_cmd

_log = fancylogger.getLogger('easyblocks.generic.systemcompiler')

KNOWN_SYSTEM_COMPILERS = {
    'GCC': 'gcc',
    'GCCcore': 'gcc',
    'icc': 'icc',
    'ifort': 'ifort',
}


def extract_compiler_version(compiler_name):
    """Extract compiler version for provided compiler_name."""
    # look for 3-4 digit version number, surrounded by spaces
    # examples:
    # gcc (GCC) 4.4.7 20120313 (Red Hat 4.4.7-11)
    # Intel(R) C Intel(R) 64 Compiler XE for applications running on Intel(R) 64, Version 15.0.1.133 Build 20141023
    version_regex = re.compile(r'\s([0-9]+(?:\.[0-9]+){1,3})\s', re.M)
    if compiler_name == 'gcc':
        res = run_shell_cmd("gcc --version")
        out = res.output
        res = version_regex.search(out)
        if res is None:
            raise EasyBuildError("Could not extract GCC version from %s", out)
        compiler_version = res.group(1)
    elif compiler_name in ['icc', 'ifort']:
        # A fully resolved icc/ifort (without symlinks) includes the release version in the path
        # e.g. .../composer_xe_2015.3.187/bin/intel64/icc
        # Match the last incidence of _ since we don't know what might be in the path, then split it up on /
        compiler_path = which(compiler_name)
        if compiler_path:
            compiler_version = resolve_path(compiler_path).split('_')[-1].split('/')[0]
        else:
            raise EasyBuildError("Compiler command '%s' not found", compiler_name)
        # Check what we have looks like a version number (the regex we use requires spaces around the version number)
        if version_regex.search(' ' + compiler_version + ' ') is None:
            error_msg = f"Derived Intel compiler version '{compiler_version}' doesn't look correct, "
            error_msg += "is compiler installed in a path like '.../composer_xe_0000.0.000/bin/intel64/icc'?"
            raise EasyBuildError(error_msg)
    else:
        raise EasyBuildError("Unknown compiler %s", compiler_name)

    if compiler_version:
        _log.debug("Extracted compiler version '%s' for %s", compiler_version, compiler_name)
    else:
        raise EasyBuildError("Failed to extract compiler version for %s using regex pattern '%s' from: %s",
                             compiler_name, version_regex.pattern, out)

    return compiler_version


class SystemCompiler(Bundle):
    """
    Support for generating a module file for the system compiler with specified name.

    The compiler is expected to be available in $PATH, required libraries are assumed to be readily available.

    Specifying 'system' as a version leads to using the derived compiler version in the generated module;
    if an actual version is specified, it is checked against the derived version of the system compiler that was found.
    """

    @staticmethod
    def extra_options():
        """Add custom easyconfig parameters for SystemCompiler easyblock."""
        # Gather extra_vars from inherited classes, order matters here to make this work without problems in __init__
        extra_vars = Bundle.extra_options()
        # Add an option to add all module path extensions to the resulting easyconfig
        # This is useful if you are importing a compiler from a non-default path
        extra_vars.update({
            'generate_standalone_module': [
                False,
                "Add known path/library extensions and environment variables for the compiler to the final module",
                CUSTOM
            ],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Extra initialization: keep track of values that may change due to modifications to the version."""
        super().__init__(*args, **kwargs)

        # by default rely on Bundle easyblock, child SytemCompiler subclasses
        # might define other easyblocks to handle standalone modules
        self.compiler_class = Bundle

        # Keep track of original values of vars that are subject to change, for restoring later.
        # The version is determined/matched from the installation and the installdir is determined from the system
        # (the original is used to store the EB logs)
        self.orig_version = self.cfg['version']
        self.orig_installdir = self.installdir

    def prepare_step(self, *args, **kwargs):
        """Do compiler appropriate prepare step, determine system compiler version and prefix."""
        # disable RPATH as SystemCompiler just generates wrapper modules
        self.toolchain.options['rpath'] = False

        self.compiler_class.prepare_step(self, *args, **kwargs)

        # Determine compiler path (real path, with resolved symlinks)
        try:
            compiler_name = KNOWN_SYSTEM_COMPILERS[self.cfg['name']]
        except KeyError as err:
            raise EasyBuildError(f"EasyConfig '{self.cfg['name']}' has no known system compilers") from err

        path_to_compiler = which(compiler_name)
        if path_to_compiler:
            path_to_compiler = resolve_path(path_to_compiler)
            self.log.info(f"Found path to compiler '{compiler_name}' (with symlinks resolved): {path_to_compiler}")
        else:
            raise EasyBuildError(f"{compiler_name} not found in $PATH")

        # Determine compiler version
        self.compiler_version = extract_compiler_version(compiler_name)

        # Determine installation prefix
        if compiler_name == 'gcc':
            # strip off 'bin/gcc'
            self.compiler_prefix = os.path.dirname(os.path.dirname(path_to_compiler))

        elif compiler_name in ['icc', 'ifort']:
            intelvars_fn = path_to_compiler + 'vars.sh'
            if os.path.isfile(intelvars_fn):
                self.log.debug("Trying to determine compiler install prefix from %s", intelvars_fn)
                intelvars_txt = read_file(intelvars_fn)
                prod_dir_regex = re.compile(r'^PROD_DIR=(.*)$', re.M)
                res = prod_dir_regex.search(intelvars_txt)
                if res:
                    self.compiler_prefix = res.group(1)
                else:
                    raise EasyBuildError("Failed to determine %s installation prefix from %s",
                                         compiler_name, intelvars_fn)
            else:
                # strip off 'bin/intel*/icc'
                self.compiler_prefix = os.path.dirname(os.path.dirname(os.path.dirname(path_to_compiler)))

            # For versions 2016+ of Intel compilers they changed the installation path so must shave off 2 more
            # directories from result of the above
            if LooseVersion(self.compiler_version) >= LooseVersion('2016'):
                self.compiler_prefix = os.path.dirname(os.path.dirname(self.compiler_prefix))

        else:
            raise EasyBuildError(f"Unknown system compiler {self.cfg['name']}")

        if not os.path.exists(self.compiler_prefix):
            raise EasyBuildError(
                f"Prefix path derived for system compiler ({compiler_name}) does not exist: {self.compiler_prefix}!"
            )
        self.log.debug(
            f"Derived version/install prefix for system compiler {compiler_name}: "
            f"{self.compiler_version}, {self.compiler_prefix}"
        )

        # If EasyConfig specified "real" version (not 'system' which means 'derive automatically'), check it
        if self.cfg['version'] == 'system':
            self.log.info(
                f"Found requested version '{self.cfg['version']}', derived from compiler as '{self.compiler_version}'"
            )
        elif self.cfg['version'] != self.compiler_version:
            raise EasyBuildError(
                f"Requested version ({self.cfg['version']}) does not match version "
                f"reported by compiler ({self.compiler_version})"
            )

    def make_installdir(self, dontcreate=None):
        """Custom implementation of make installdir: do nothing, do not touch system compiler directories and files."""
        pass

    def make_module_step(self, fake=False):
        """
        Custom module step for SystemCompiler: make 'EBROOT' and 'EBVERSION' reflect actual system compiler version
        and install path.
        """
        # For module file generation: temporarly set version and installdir to system compiler values
        self.cfg['version'] = self.compiler_version
        self.installdir = self.compiler_prefix

        # Generate module
        module_generator_class = Bundle
        if self.compiler_class is not None:
            if self.compiler_prefix in ['/usr', '/usr/local']:
                # reset module load environment for system compilers in default $PATHS
                # as such a module would be a potential shell killer
                print_warning(
                    "Ignoring option 'generate_standalone_module' since installation prefix is "
                    f"already in $PATH: {self.compiler_prefix}"
                )
                module_vars = [str(env_var) for env_var in self.module_load_environment]
                for env_var in module_vars:
                    self.module_load_environment.remove(env_var)
            else:
                # rely on compiler class module step to generate standalone module
                module_generator_class = self.compiler_class

        module_path = module_generator_class.make_module_step(self, fake=fake)

        # Reset version and installdir to EasyBuild values
        self.installdir = self.orig_installdir
        self.cfg['version'] = self.orig_version

        return module_path

    def make_module_extend_modpath(self):
        """
        Custom prepend-path statements for extending $MODULEPATH: use version specified in easyconfig file (e.g.,
        "system") rather than the actual version (e.g., "4.8.2").
        """
        # temporarly set switch back to version specified in easyconfig file (e.g., "system")
        self.cfg['version'] = self.orig_version

        # Retrieve module path extensions
        extended_modpath = self.compiler_class.make_module_extend_modpath(self)

        # Reset to actual compiler version (e.g., "4.8.2")
        self.cfg['version'] = self.compiler_version

        return extended_modpath

    def make_module_extra(self, *args, **kwargs):
        """Add any additional module text."""
        return self.compiler_class.make_module_extra(self, *args, **kwargs)

    def post_processing_step(self, *args, **kwargs):
        """Do nothing."""
        pass

    def cleanup_step(self):
        """Do nothing."""
        pass

    def permissions_step(self):
        """Do nothing."""
        pass

    def sanity_check_step(self):
        """
        Nothing is being installed, so just being able to load the (fake) module is sufficient
        """
        self.log.info(f"Testing loading of module '{self.full_mod_name}' by means of sanity check")
        fake_mod_data = self.load_fake_module(purge=True)
        self.log.debug("Cleaning up after testing loading of module")
        self.clean_up_fake_module(fake_mod_data)



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Ansys Electronics Desktop

@author: Chia-Jung Hsu (Chalmers University of Technology)
"""
import os
import glob
import shutil
import tempfile

from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_shell_cmd


class EB_AEDT(PackedBinary):
    """Support for installing Ansys Electronics Desktop."""

    def __init__(self, *args, **kwargs):
        """Initialize Ansys Electronics Desktop specific variables."""
        super().__init__(*args, **kwargs)
        self.subdir = None

    def _set_subdir(self):

        ver = LooseVersion(self.version)
        if ver > LooseVersion("2024R1"):
            pattern = 'v*/AnsysEM/'
        else:
            pattern = 'v*/Linux*/'
        idirs = glob.glob(os.path.join(self.installdir, pattern))

        if len(idirs) == 1:
            self.subdir = os.path.relpath(idirs[0], self.installdir)
        else:
            raise EasyBuildError(f"Failed to locate single install subdirectory {pattern}")

    def install_step(self):
        """Install Ansys Electronics Desktop using its setup tool."""

        ver = LooseVersion(self.version)
        licserv = os.getenv('EB_AEDT_LICENSE_SERVER')
        licport = os.getenv('EB_AEDT_LICENSE_SERVER_PORT')
        licservs = licserv.split(',')
        licservopt = ["-DLICENSE_SERVER%d=%s" % (i, serv) for i, serv in enumerate(licservs, 1)]
        redundant = len(licservs) > 1

        tempdir = tempfile.TemporaryDirectory()
        if ver > LooseVersion("2024R1"):
            installer = "./INSTALL"
            cmd = " ".join([
                installer,
                "-silent",
                "-install_dir %s" % self.installdir,
                "-licserverinfo :%s:%s" % (licport, licserv),
                "-usetempdir %s" % tempdir.name,
            ])
        else:
            installer = "./Linux/AnsysEM/Disk1/InstData/setup.exe"

            cmd = " ".join([
                installer,
                "-i silent",
                "-DUSER_INSTALL_DIR=%s" % self.installdir,
                "-DTMP_DIR=%s" % tempdir.name,
                "-DLIBRARY_COMMON_INSTALL=1",
                "-DSPECIFY_LIC_CFG=1",
                "-DREDUNDANT_SERVERS=%d" % redundant,
                *licservopt,
                "-DSPECIFY_PORT=1",
                "-DLICENSE_PORT=%s" % licport,
            ])
        run_shell_cmd(cmd)

    def post_processing_step(self):
        """Disable OS check and set LC_ALL/LANG for runtime"""
        if not self.subdir:
            self._set_subdir()

        # Clean script file to disable over restrict OS checking
        with open(os.path.join(self.installdir, self.subdir, "VerifyOS.bash"), "w") as f:
            f.write("")

        # Follow the settings in .setup_runtime and .setup_runtime_mpi to set LC_ALL and LANG
        # If LC_ALL is not set properly, AEDT throws the following runtime error:
        # what():  locale::facet::_S_create_c_locale name not valid
        for fname in [".setup_runtime", ".setup_runtime_mpi"]:
            with open(os.path.join(self.installdir, self.subdir, fname), "r") as f:
                orig = f.read()
            with open(os.path.join(self.installdir, self.subdir, fname), "w") as f:
                f.write("export LC_ALL=C\n")
                f.write("export LANG=C\n")
                f.write(orig)

    def make_module_extra(self):
        """Extra module entries for Ansys Electronics Desktop."""

        txt = super().make_module_extra()
        ver = LooseVersion(self.version)
        short_ver = self.version[2:].replace('R', '')

        if ver > LooseVersion("2024R1"):
            pattern = 'v*/AnsysEM/'
        else:
            pattern = 'v*/Linux*/'
        idirs = glob.glob(os.path.join(self.installdir, pattern))

        if len(idirs) == 1:
            subdir = os.path.relpath(idirs[0], self.installdir)
            # PyAEDT and other tools use the variable to find available AEDT versions
            txt += self.module_generator.set_environment('ANSYSEM_ROOT%s' % short_ver,
                                                         os.path.join(self.installdir, subdir))

            txt += self.module_generator.prepend_paths('PATH', subdir)
            txt += self.module_generator.prepend_paths('LD_LIBRARY_PATH', subdir)
            txt += self.module_generator.prepend_paths('LIBRARY_PATH', subdir)

        return txt

    def sanity_check_step(self):
        """Custom sanity check for Ansys Electronics Desktop."""
        if not self.subdir:
            self._set_subdir()

        custom_paths = {
            'files': [os.path.join(self.subdir, 'ansysedt')],
            'dirs': [self.subdir],
        }

        # Since 2025R1, test examples cannot be directly found in builddir
        # Copy a test example from installdir to a tempdir to prevent writing additional file into installdir
        inpfilesrc = os.path.join(self.installdir, self.subdir, 'Examples', 'RMxprt', 'lssm', 'sm-1.aedt')
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(inpfilesrc, tempdir)
            inpfile = os.path.join(tempdir, 'sm-1.aedt')
            custom_commands = ['ansysedt -ng -batchsolve -Distributed -monitor %s' % inpfile]

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for Go packages, implemented as an EasyBlock

@author: Pavel Grochal (INUITS)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os
from easybuild.tools import LooseVersion

import easybuild.tools.environment as env
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.systemtools import AARCH32, AARCH64, X86_64, get_cpu_architecture
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.toolchain.compiler import OPTARCH_GENERIC


class GoPackage(EasyBlock):
    """Builds and installs a Go package, and provides a dedicated module file."""

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to Go packages."""
        if extra_vars is None:
            extra_vars = {}
        extra_vars.update({
            'modulename': [None, "Module name of the Go package, when building non-native module", CUSTOM],
            'forced_deps': [None, "Force specific version of Go package, when building non-native module", CUSTOM],
        })
        return extra_vars

    def go_microarch_opt(self):
        """
        Microarchitecture optimization support
        see https://go.dev/wiki/MinimumRequirements#architectures
        """
        if LooseVersion(get_software_version('Go')) < LooseVersion("1.18"):
            self.log.debug("Go version %s does not support microarchitecture optimization", self.version)
            return None

        optarch = build_option('optarch') or ''
        if isinstance(optarch, dict):
            # optarch specified by compiler family
            optarch = optarch.get(self.toolchain.comp_family(), '')
        optarch = optarch.upper()

        microarch = None
        opt_level = None

        if get_cpu_architecture() == AARCH32:
            # default to VFPv1, default if cross compiling
            microarch = "GOARM"
            opt_level = "6"
            if 'VFPV3' in optarch:
                # use VFPv3; usually Cortex-A cores
                opt_level = "7"
            elif 'VFP' in optarch:
                # use VFPv1 only; usually ARM11 or better (VFPv2 or better is also supported)
                opt_level = "6"
            elif optarch == OPTARCH_GENERIC:
                opt_level = "5"
            elif optarch:
                opt_level = None

        elif get_cpu_architecture() == AARCH64:
            if LooseVersion(get_software_version('Go')) < LooseVersion("1.23"):
                self.log.debug(
                    "Go version %s does not support microarchitecture optimization for %s", self.version, AARCH64
                )
                return None

            # default to v8.0, default in Go for AARCH64
            microarch = "GOARM64"
            opt_level = "v8.0"
            # Allowed values are v8.{0-9} and v9.{0-5}. This may be followed by an option specifying extensions
            # implemented by target hardware. Example: GOARM64=v8.0,lse
            if optarch.startswith("v8.") or optarch.startswith("v9."):
                opt_level = optarch
            elif optarch == OPTARCH_GENERIC:
                opt_level = "v8.0"
            elif optarch:
                opt_level = None

        elif get_cpu_architecture() == X86_64:
            # default to x86-64-v3 with AVX2
            microarch = "GOAMD64"
            opt_level = "v3"
            if optarch.endswith("V4") or 'AVX512' in optarch:
                opt_level = "v4"
            elif optarch.endswith("V3") or 'AVX' in optarch:
                opt_level = "v3"
            elif optarch.endswith("V2") or 'SSE4' in optarch or 'SSE3' in optarch:
                opt_level = "v2"
            elif optarch == OPTARCH_GENERIC or optarch.endswith("V1"):
                opt_level = "v1"
            elif optarch:
                opt_level = None

        if microarch and opt_level is None:
            warn_msg = f"Unknown optarch setting given for Go architecture ({microarch}): {optarch}."
            warn_msg += f" Building '{self.name}' with auto-detection in Go."
            self.log.warning(warn_msg)
            print_warning(warn_msg)
            return None

        self.log.debug(f"Setting 'optarch={optarch}' converted to Go environment variable '{microarch}:{opt_level}'")
        return (microarch, opt_level)

    def prepare_step(self, *args, **kwargs):
        """Go-specific preparations."""
        super().prepare_step(*args, **kwargs)

        if get_software_root('Go') is None:
            raise EasyBuildError("Failed to pick go command to use. Is it listed in dependencies?")

        if LooseVersion(get_software_version('Go')) < LooseVersion("1.11"):
            raise EasyBuildError("Go version < 1.11 doesn't support installing modules from go.mod")

    def configure_step(self):
        """Configure Go package build/install."""

        # Move compiled .a files into builddir, else they pollute $HOME/go
        env.setvar('GOPATH', self.builddir, verbose=False)
        # enforce use of go modules
        env.setvar('GO111MODULE', 'on', verbose=False)
        # set bin folder
        env.setvar('GOBIN', os.path.join(self.installdir, 'bin'), verbose=False)
        # microarchitecture
        microarch = self.go_microarch_opt()
        if microarch is not None:
            env.setvar(*microarch, verbose=False)

        # creates log entries for go being used, for debugging
        run_shell_cmd("go version", hidden=True)
        run_shell_cmd("go env", hidden=True)

    def build_step(self):
        """If Go package is not native go module, lets try to make the module."""

        go_mod_file = 'go.mod'
        go_sum_file = 'go.sum'

        if not os.path.exists(go_mod_file) or not os.path.isfile(go_mod_file):
            self.log.warning("go.mod not found! This is not natively supported go module. Trying to init module.")

            if self.cfg['modulename'] is None:
                raise EasyBuildError("Installing non-native go module. You need to specify 'modulename' in easyconfig")

            # for more information about migrating to go modules
            # see: https://blog.golang.org/migrating-to-go-modules

            # go mod init
            cmd = ' '.join(['go', 'mod', 'init', self.cfg['modulename']])
            run_shell_cmd(cmd)

            if self.cfg['forced_deps']:
                for dep in self.cfg['forced_deps']:
                    # go get specific dependencies which locks them in go.mod
                    cmd = ' '.join(['go', 'get', '%s@%s' % dep])
                    run_shell_cmd(cmd)

            # note: ... (tripledot) used below is not a typo, but go wildcard pattern
            # which means: anything you can find in this directory, including all subdirectories
            # see: 'go help packages' or https://golang.org/pkg/cmd/go/internal/help/
            # see: https://stackoverflow.com/a/28031651/2047157

            # building and testing will add packages to go.mod
            run_shell_cmd('go build ./...')
            run_shell_cmd('go test ./...')

            # tidy up go.mod
            run_shell_cmd('go mod tidy')

            # build and test again, to ensure go mod tidy didn't removed anything needed
            run_shell_cmd('go build ./...')
            run_shell_cmd('go test ./...')

            self.log.warning('Include generated go.mod and go.sum via patch to ensure locked dependencies '
                             'and run this easyconfig again.')
            run_shell_cmd('cat go.mod')
            run_shell_cmd('cat go.sum')

        if not os.path.exists(go_sum_file) or not os.path.isfile(go_sum_file):
            raise EasyBuildError("go.sum not found! This module has no locked dependency versions.")

    def install_step(self):
        """Install Go package to a custom path"""

        # actually install Go package
        cmd = ' '.join([
            self.cfg['preinstallopts'],
            'go',
            'install',
            # print commands as they are executed,
            # including downloading and installing of package deps as listed in the go.mod file
            '-x',
            self.cfg['installopts'],
        ])
        run_shell_cmd(cmd)

    def sanity_check_step(self):
        """Custom sanity check for Go package."""

        # Check if GoPackage produced binary and can run help on it
        custom_paths = {
            'files': ['bin/%s' % self.name.lower()],
            'dirs': [],
        }
        custom_commands = ['%s --help' % self.name.lower()]

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def sanity_check_rpath(self, rpath_dirs=None):
        super().sanity_check_rpath(rpath_dirs=rpath_dirs, check_readelf_rpath=False)



# #
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for install the Intel C/C++ compiler suite, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Fokko Masselink
"""

import os
import re
from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.intelbase import IntelBase, COMP_ALL
from easybuild.easyblocks.tbb import get_tbb_gccprefix
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import MODULE_LOAD_ENV_HEADERS
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


def get_icc_version():
    """Obtain icc version string via 'icc --version'."""
    cmd = "icc --version"
    res = run_shell_cmd(cmd)

    ver_re = re.compile(r"^icc \(ICC\) (?P<version>[0-9.]+) [0-9]+$", re.M)
    version = ver_re.search(res.output).group('version')

    return version


class EB_icc(IntelBase):
    """
    Support for installing icc
    - minimum version suported: 2020.0
    """

    def __init__(self, *args, **kwargs):
        """Constructor, initialize class variables."""
        super().__init__(*args, **kwargs)

        self.comp_libs_subdir = ""

        # need to make sure version is an actual version
        # required because of support in SystemCompiler generic easyblock to specify 'system' as version,
        # which results in deriving the actual compiler version
        # comparing a non-version like 'system' with an actual version like '2016' fails with TypeError in Python 3.x
        if re.match(r'^[0-9]+.*', self.version):

            if LooseVersion(self.version) < LooseVersion('2020'):
                raise EasyBuildError(
                    f"Version {self.version} of {self.name} is unsupported. Mininum supported version is 2020.0."
                )

            self.comp_libs_subdir = os.path.join(f'compilers_and_libraries_{self.version}', 'linux')

            if self.cfg['components'] is None:
                # we need to use 'ALL' by default,
                # using 'DEFAULTS' results in key things not being installed (e.g. bin/icc)
                self.cfg['components'] = [COMP_ALL]
                self.log.debug(
                    f"Missing components specification, required for version {self.version}. "
                    f"Using {self.cfg['components']} instead."
                )

        # define list of subdirectories in search paths of module load environment
        # some of these paths only apply to certain versions, but that doesn't really matter
        # existence of paths is checked by module generator before 'prepend-paths' statements are included

        # new directory layout since Intel Parallel Studio XE 2016
        # https://software.intel.com/en-us/articles/new-directory-layout-for-intel-parallel-studio-xe-2016
        # in recent Intel compiler distributions, the actual binaries are
        # in deeper directories, and symlinked in top-level directories
        # however, not all binaries are symlinked (e.g. mcpcom is not)
        # we only need to include the deeper directories (same as compilervars.sh)
        def comp_libs_subdir_paths(*subdir_paths):
            """Utility method to prepend self.comp_libs_subdir to list of paths"""
            try:
                return [os.path.join(self.comp_libs_subdir, path) for path in subdir_paths]
            except TypeError:
                raise EasyBuildError(f"Cannot prepend {self.comp_libs_subdir} to {subdir_paths}: wrong data type")

        self.module_load_environment.PATH = comp_libs_subdir_paths(
            'bin/intel64',
            'ipp/bin/intel64',
            'mpi/intel64/bin',
            'tbb/bin/emt64',
            'tbb/bin/intel64',
        )
        # in the end we set 'LIBRARY_PATH' equal to 'LD_LIBRARY_PATH'
        self.module_load_environment.LD_LIBRARY_PATH = comp_libs_subdir_paths(
            'lib',
            'compiler/lib/intel64',
            'debugger/ipt/intel64/lib',
            'ipp/lib/intel64',
            'mkl/lib/intel64',
            'mpi/intel64',
            f"tbb/lib/intel64/{get_tbb_gccprefix(os.path.join(self.installdir, 'tbb/lib/intel64'))}",
        )
        # 'include' is deliberately omitted, including it causes problems, e.g. with complex.h and std::complex
        # cfr. https://software.intel.com/en-us/forums/intel-c-compiler/topic/338378
        self.module_load_environment.set_alias_vars(MODULE_LOAD_ENV_HEADERS, comp_libs_subdir_paths(
            'daal/include',
            'ipp/include',
            'mkl/include',
            'tbb/include',
        ))
        self.module_load_environment.CLASSPATH = comp_libs_subdir_paths('daal/lib/daal.jar')
        self.module_load_environment.DAALROOT = comp_libs_subdir_paths('daal')
        self.module_load_environment.IPPROOT = comp_libs_subdir_paths('ipp')
        self.module_load_environment.MANPATH = comp_libs_subdir_paths(
            'debugger/gdb/intel64/share/man',
            'man/common',
            'man/en_US',
            'share/man',
        )
        self.module_load_environment.TBBROOT = comp_libs_subdir_paths('tbb')

        # Debugger requires INTEL_PYTHONHOME, which only allows for a single value
        self.debuggerpath = f"debugger_{self.version.split('.')[0]}"
        self.module_load_environment.LD_LIBRARY_PATH.extend([
            os.path.join(self.debuggerpath, 'libipt/intel64/lib'),
            'daal/lib/intel64_lin',  # out of self.comp_libs_subdir to not break libipt library loading for gdb
        ])
        self.module_load_environment.PATH.append(
            os.path.join(self.debuggerpath, 'gdb', 'intel64', 'bin')
        )

        # 'lib/intel64' is deliberately listed last, so it gets precedence over subdirs
        self.module_load_environment.LD_LIBRARY_PATH.extend(comp_libs_subdir_paths('lib/intel64'))

        self.module_load_environment.LIBRARY_PATH = self.module_load_environment.LD_LIBRARY_PATH

    def sanity_check_step(self):
        """Custom sanity check paths for icc."""

        binprefix = 'bin'
        binfiles = ['icc', 'icpc']
        binaries = [os.path.join(binprefix, f) for f in binfiles]

        libprefix = 'lib/intel64'
        libraries = [os.path.join(libprefix, f'lib{lib}') for lib in ['iomp5.a', f'iomp5.{get_shared_lib_ext()}']]

        headers = ['include/omp.h']

        custom_paths = {
            'files': binaries + libraries + headers,
            'dirs': [],
        }

        # make very sure that expected 'compilers_and_libraries_<VERSION>/linux' subdir is there for recent versions,
        # since we rely on it being there in module_load_environment
        if self.comp_libs_subdir:
            custom_paths['dirs'].append(self.comp_libs_subdir)

        custom_commands = ["which icc"]

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_step(self, *args, **kwargs):
        """
        Additional paths for module load environment that are conditional
        """
        # only set $IDB_HOME if idb exists
        idb_home_subdir = 'bin/intel64'
        if os.path.isfile(os.path.join(self.installdir, idb_home_subdir, 'idb')):
            self.module_load_environment.IDB_HOME = [idb_home_subdir]

        return super().make_module_step(*args, **kwargs)

    def make_module_extra(self, *args, **kwargs):
        """Additional custom variables for icc: $INTEL_PYTHONHOME."""
        txt = super().make_module_extra(*args, **kwargs)

        intel_pythonhome = os.path.join(self.installdir, self.debuggerpath, 'python', 'intel64')
        if os.path.isdir(intel_pythonhome):
            txt += self.module_generator.set_environment('INTEL_PYTHONHOME', intel_pythonhome)

        # on Debian/Ubuntu, /usr/include/x86_64-linux-gnu needs to be included in $CPATH for icc
        res = run_shell_cmd("gcc -print-multiarch", fail_on_error=False, hidden=True)
        multiarch_inc_subdir = res.output.strip()
        if res.exit_code == 0 and multiarch_inc_subdir:
            multiarch_inc_dir = os.path.join('/usr', 'include', multiarch_inc_subdir)
            for envar in self.module_load_environment.alias_vars(MODULE_LOAD_ENV_HEADERS):
                self.log.info(f"Adding multiarch include path '{multiarch_inc_dir}' to ${envar} in generated module")
                # system location must be appended at the end, so use append_paths
                txt += self.module_generator.append_paths(envar, [multiarch_inc_dir], allow_abs=True)

        return txt
